from enum import StrEnum


class Icons(StrEnum):
    ADD = "\uea60"
    PLUS = "\uea60"
    GIST_NEW = "\uea60"
    REPO_CREATE = "\uea60"
    LIGHTBULB = "\uea61"
    LIGHT_BULB = "\uea61"
    REPO = "\uea62"
    REPO_DELETE = "\uea62"
    GIST_FORK = "\uea63"
    REPO_FORKED = "\uea63"
    GIT_PULL_REQUEST = "\uea64"
    GIT_PULL_REQUEST_ABANDONED = "\uea64"
    RECORD_KEYS = "\uea65"
    KEYBOARD = "\uea65"
    TAG = "\uea66"
    GIT_PULL_REQUEST_LABEL = "\uea66"
    TAG_ADD = "\uea66"
    TAG_REMOVE = "\uea66"
    PERSON = "\uea67"
    PERSON_FOLLOW = "\uea67"
    PERSON_OUTLINE = "\uea67"
    PERSON_FILLED = "\uea67"
    GIT_BRANCH = "\uea68"
    GIT_BRANCH_CREATE = "\uea68"
    GIT_BRANCH_DELETE = "\uea68"
    SOURCE_CONTROL = "\uea68"
    MIRROR = "\uea69"
    MIRROR_PUBLIC = "\uea69"
    STAR = "\uea6a"
    STAR_ADD = "\uea6a"
    STAR_DELETE = "\uea6a"
    STAR_EMPTY = "\uea6a"
    COMMENT = "\uea6b"
    COMMENT_ADD = "\uea6b"
    ALERT = "\uea6c"
    WARNING = "\uea6c"
    SEARCH = "\uea6d"
    SEARCH_SAVE = "\uea6d"
    LOG_OUT = "\uea6e"
    SIGN_OUT = "\uea6e"
    LOG_IN = "\uea6f"
    SIGN_IN = "\uea6f"
    EYE = "\uea70"
    EYE_UNWATCH = "\uea70"
    EYE_WATCH = "\uea70"
    CIRCLE_FILLED = "\uea71"
    PRIMITIVE_DOT = "\uea71"
    CLOSE_DIRTY = "\uea71"
    DEBUG_BREAKPOINT = "\uea71"
    DEBUG_BREAKPOINT_DISABLED = "\uea71"
    DEBUG_HINT = "\uea71"
    TERMINAL_DECORATION_SUCCESS = "\uea71"
    PRIMITIVE_SQUARE = "\uea72"
    EDIT = "\uea73"
    PENCIL = "\uea73"
    INFO = "\uea74"
    ISSUE_OPENED = "\uea74"
    GIST_PRIVATE = "\uea75"
    GIT_FORK_PRIVATE = "\uea75"
    LOCK = "\uea75"
    MIRROR_PRIVATE = "\uea75"
    CLOSE = "\uea76"
    REMOVE_CLOSE = "\uea76"
    X = "\uea76"
    REPO_SYNC = "\uea77"
    SYNC = "\uea77"
    CLONE = "\uea78"
    DESKTOP_DOWNLOAD = "\uea78"
    BEAKER = "\uea79"
    MICROSCOPE = "\uea79"
    VM = "\uea7a"
    DEVICE_DESKTOP = "\uea7a"
    FILE = "\uea7b"
    FILE_TEXT = "\uea7b"
    MORE = "\uea7c"
    ELLIPSIS = "\uea7c"
    KEBAB_HORIZONTAL = "\uea7c"
    MAIL_REPLY = "\uea7d"
    REPLY = "\uea7d"
    ORGANIZATION = "\uea7e"
    ORGANIZATION_FILLED = "\uea7e"
    ORGANIZATION_OUTLINE = "\uea7e"
    NEW_FILE = "\uea7f"
    FILE_ADD = "\uea7f"
    NEW_FOLDER = "\uea80"
    FILE_DIRECTORY_CREATE = "\uea80"
    TRASH = "\uea81"
    TRASHCAN = "\uea81"
    HISTORY = "\uea82"
    CLOCK = "\uea82"
    FOLDER = "\uea83"
    FILE_DIRECTORY = "\uea83"
    SYMBOL_FOLDER = "\uea83"
    LOGO_GITHUB = "\uea84"
    MARK_GITHUB = "\uea84"
    GITHUB = "\uea84"
    TERMINAL = "\uea85"
    CONSOLE = "\uea85"
    REPL = "\uea85"
    ZAP = "\uea86"
    SYMBOL_EVENT = "\uea86"
    ERROR = "\uea87"
    STOP = "\uea87"
    VARIABLE = "\uea88"
    SYMBOL_VARIABLE = "\uea88"
    ARRAY = "\uea8a"
    SYMBOL_ARRAY = "\uea8a"
    SYMBOL_MODULE = "\uea8b"
    SYMBOL_PACKAGE = "\uea8b"
    SYMBOL_NAMESPACE = "\uea8b"
    SYMBOL_OBJECT = "\uea8b"
    SYMBOL_METHOD = "\uea8c"
    SYMBOL_FUNCTION = "\uea8c"
    SYMBOL_CONSTRUCTOR = "\uea8c"
    SYMBOL_BOOLEAN = "\uea8f"
    SYMBOL_NULL = "\uea8f"
    SYMBOL_NUMERIC = "\uea90"
    SYMBOL_NUMBER = "\uea90"
    SYMBOL_STRUCTURE = "\uea91"
    SYMBOL_STRUCT = "\uea91"
    SYMBOL_PARAMETER = "\uea92"
    SYMBOL_TYPE_PARAMETER = "\uea92"
    SYMBOL_KEY = "\uea93"
    SYMBOL_TEXT = "\uea93"
    SYMBOL_REFERENCE = "\uea94"
    GO_TO_FILE = "\uea94"
    SYMBOL_ENUM = "\uea95"
    SYMBOL_VALUE = "\uea95"
    SYMBOL_RULER = "\uea96"
    SYMBOL_UNIT = "\uea96"
    ACTIVATE_BREAKPOINTS = "\uea97"
    ARCHIVE = "\uea98"
    ARROW_BOTH = "\uea99"
    ARROW_DOWN = "\uea9a"
    ARROW_LEFT = "\uea9b"
    ARROW_RIGHT = "\uea9c"
    ARROW_SMALL_DOWN = "\uea9d"
    ARROW_SMALL_LEFT = "\uea9e"
    ARROW_SMALL_RIGHT = "\uea9f"
    ARROW_SMALL_UP = "\ueaa0"
    ARROW_UP = "\ueaa1"
    BELL = "\ueaa2"
    BOLD = "\ueaa3"
    BOOK = "\ueaa4"
    BOOKMARK = "\ueaa5"
    DEBUG_BREAKPOINT_CONDITIONAL_UNVERIFIED = "\ueaa6"
    DEBUG_BREAKPOINT_CONDITIONAL = "\ueaa7"
    DEBUG_BREAKPOINT_CONDITIONAL_DISABLED = "\ueaa7"
    DEBUG_BREAKPOINT_DATA_UNVERIFIED = "\ueaa8"
    DEBUG_BREAKPOINT_DATA = "\ueaa9"
    DEBUG_BREAKPOINT_DATA_DISABLED = "\ueaa9"
    DEBUG_BREAKPOINT_LOG_UNVERIFIED = "\ueaaa"
    DEBUG_BREAKPOINT_LOG = "\ueaab"
    DEBUG_BREAKPOINT_LOG_DISABLED = "\ueaab"
    BRIEFCASE = "\ueaac"
    BROADCAST = "\ueaad"
    BROWSER = "\ueaae"
    BUG = "\ueaaf"
    CALENDAR = "\ueab0"
    CASE_SENSITIVE = "\ueab1"
    CHECK = "\ueab2"
    CHECKLIST = "\ueab3"
    CHEVRON_DOWN = "\ueab4"
    CHEVRON_LEFT = "\ueab5"
    CHEVRON_RIGHT = "\ueab6"
    CHEVRON_UP = "\ueab7"
    CHROME_CLOSE = "\ueab8"
    CHROME_MAXIMIZE = "\ueab9"
    CHROME_MINIMIZE = "\ueaba"
    CHROME_RESTORE = "\ueabb"
    CIRCLE_OUTLINE = "\ueabc"
    CIRCLE = "\ueabc"
    DEBUG_BREAKPOINT_UNVERIFIED = "\ueabc"
    TERMINAL_DECORATION_INCOMPLETE = "\ueabc"
    CIRCLE_SLASH = "\ueabd"
    CIRCUIT_BOARD = "\ueabe"
    CLEAR_ALL = "\ueabf"
    CLIPPY = "\ueac0"
    CLOSE_ALL = "\ueac1"
    CLOUD_DOWNLOAD = "\ueac2"
    CLOUD_UPLOAD = "\ueac3"
    CODE = "\ueac4"
    COLLAPSE_ALL = "\ueac5"
    COLOR_MODE = "\ueac6"
    COMMENT_DISCUSSION = "\ueac7"
    CREDIT_CARD = "\ueac9"
    DASH = "\ueacc"
    DASHBOARD = "\ueacd"
    DATABASE = "\ueace"
    DEBUG_CONTINUE = "\ueacf"
    DEBUG_DISCONNECT = "\uead0"
    DEBUG_PAUSE = "\uead1"
    DEBUG_RESTART = "\uead2"
    DEBUG_START = "\uead3"
    DEBUG_STEP_INTO = "\uead4"
    DEBUG_STEP_OUT = "\uead5"
    DEBUG_STEP_OVER = "\uead6"
    DEBUG_STOP = "\uead7"
    DEBUG = "\uead8"
    DEVICE_CAMERA_VIDEO = "\uead9"
    DEVICE_CAMERA = "\ueada"
    DEVICE_MOBILE = "\ueadb"
    DIFF_ADDED = "\ueadc"
    DIFF_IGNORED = "\ueadd"
    DIFF_MODIFIED = "\ueade"
    DIFF_REMOVED = "\ueadf"
    DIFF_RENAMED = "\ueae0"
    DIFF = "\ueae1"
    DIFF_SIDEBYSIDE = "\ueae1"
    DISCARD = "\ueae2"
    EDITOR_LAYOUT = "\ueae3"
    EMPTY_WINDOW = "\ueae4"
    EXCLUDE = "\ueae5"
    EXTENSIONS = "\ueae6"
    EYE_CLOSED = "\ueae7"
    FILE_BINARY = "\ueae8"
    FILE_CODE = "\ueae9"
    FILE_MEDIA = "\ueaea"
    FILE_PDF = "\ueaeb"
    FILE_SUBMODULE = "\ueaec"
    FILE_SYMLINK_DIRECTORY = "\ueaed"
    FILE_SYMLINK_FILE = "\ueaee"
    FILE_ZIP = "\ueaef"
    FILES = "\ueaf0"
    FILTER = "\ueaf1"
    FLAME = "\ueaf2"
    FOLD_DOWN = "\ueaf3"
    FOLD_UP = "\ueaf4"
    FOLD = "\ueaf5"
    FOLDER_ACTIVE = "\ueaf6"
    FOLDER_OPENED = "\ueaf7"
    GEAR = "\ueaf8"
    GIFT = "\ueaf9"
    GIST_SECRET = "\ueafa"
    GIST = "\ueafb"
    GIT_COMMIT = "\ueafc"
    GIT_COMPARE = "\ueafd"
    COMPARE_CHANGES = "\ueafd"
    GIT_MERGE = "\ueafe"
    GITHUB_ACTION = "\ueaff"
    GITHUB_ALT = "\ueb00"
    GLOBE = "\ueb01"
    GRABBER = "\ueb02"
    GRAPH = "\ueb03"
    GRIPPER = "\ueb04"
    HEART = "\ueb05"
    HOME = "\ueb06"
    HORIZONTAL_RULE = "\ueb07"
    HUBOT = "\ueb08"
    INBOX = "\ueb09"
    ISSUE_REOPENED = "\ueb0b"
    ISSUES = "\ueb0c"
    ITALIC = "\ueb0d"
    JERSEY = "\ueb0e"
    JSON = "\ueb0f"
    KEBAB_VERTICAL = "\ueb10"
    KEY = "\ueb11"
    LAW = "\ueb12"
    LIGHTBULB_AUTOFIX = "\ueb13"
    LINK_EXTERNAL = "\ueb14"
    LINK = "\ueb15"
    LIST_ORDERED = "\ueb16"
    LIST_UNORDERED = "\ueb17"
    LIVE_SHARE = "\ueb18"
    LOADING = "\ueb19"
    LOCATION = "\ueb1a"
    MAIL_READ = "\ueb1b"
    MAIL = "\ueb1c"
    MARKDOWN = "\ueb1d"
    MEGAPHONE = "\ueb1e"
    MENTION = "\ueb1f"
    MILESTONE = "\ueb20"
    GIT_PULL_REQUEST_MILESTONE = "\ueb20"
    MORTAR_BOARD = "\ueb21"
    MOVE = "\ueb22"
    MULTIPLE_WINDOWS = "\ueb23"
    MUTE = "\ueb24"
    NO_NEWLINE = "\ueb25"
    NOTE = "\ueb26"
    OCTOFACE = "\ueb27"
    OPEN_PREVIEW = "\ueb28"
    PACKAGE = "\ueb29"
    PAINTCAN = "\ueb2a"
    PIN = "\ueb2b"
    PLAY = "\ueb2c"
    RUN = "\ueb2c"
    PLUG = "\ueb2d"
    PRESERVE_CASE = "\ueb2e"
    PREVIEW = "\ueb2f"
    PROJECT = "\ueb30"
    PULSE = "\ueb31"
    QUESTION = "\ueb32"
    QUOTE = "\ueb33"
    RADIO_TOWER = "\ueb34"
    REACTIONS = "\ueb35"
    REFERENCES = "\ueb36"
    REFRESH = "\ueb37"
    REGEX = "\ueb38"
    REMOTE_EXPLORER = "\ueb39"
    REMOTE = "\ueb3a"
    REMOVE = "\ueb3b"
    REPLACE_ALL = "\ueb3c"
    REPLACE = "\ueb3d"
    REPO_CLONE = "\ueb3e"
    REPO_FORCE_PUSH = "\ueb3f"
    REPO_PULL = "\ueb40"
    REPO_PUSH = "\ueb41"
    REPORT = "\ueb42"
    REQUEST_CHANGES = "\ueb43"
    ROCKET = "\ueb44"
    ROOT_FOLDER_OPENED = "\ueb45"
    ROOT_FOLDER = "\ueb46"
    RSS = "\ueb47"
    RUBY = "\ueb48"
    SAVE_ALL = "\ueb49"
    SAVE_AS = "\ueb4a"
    SAVE = "\ueb4b"
    SCREEN_FULL = "\ueb4c"
    SCREEN_NORMAL = "\ueb4d"
    SEARCH_STOP = "\ueb4e"
    SERVER = "\ueb50"
    SETTINGS_GEAR = "\ueb51"
    SETTINGS = "\ueb52"
    SHIELD = "\ueb53"
    SMILEY = "\ueb54"
    SORT_PRECEDENCE = "\ueb55"
    SPLIT_HORIZONTAL = "\ueb56"
    SPLIT_VERTICAL = "\ueb57"
    SQUIRREL = "\ueb58"
    STAR_FULL = "\ueb59"
    STAR_HALF = "\ueb5a"
    SYMBOL_CLASS = "\ueb5b"
    SYMBOL_COLOR = "\ueb5c"
    SYMBOL_CONSTANT = "\ueb5d"
    SYMBOL_ENUM_MEMBER = "\ueb5e"
    SYMBOL_FIELD = "\ueb5f"
    SYMBOL_FILE = "\ueb60"
    SYMBOL_INTERFACE = "\ueb61"
    SYMBOL_KEYWORD = "\ueb62"
    SYMBOL_MISC = "\ueb63"
    SYMBOL_OPERATOR = "\ueb64"
    SYMBOL_PROPERTY = "\ueb65"
    WRENCH = "\ueb65"
    WRENCH_SUBACTION = "\ueb65"
    SYMBOL_SNIPPET = "\ueb66"
    TASKLIST = "\ueb67"
    TELESCOPE = "\ueb68"
    TEXT_SIZE = "\ueb69"
    THREE_BARS = "\ueb6a"
    THUMBSDOWN = "\ueb6b"
    THUMBSUP = "\ueb6c"
    TOOLS = "\ueb6d"
    TRIANGLE_DOWN = "\ueb6e"
    TRIANGLE_LEFT = "\ueb6f"
    TRIANGLE_RIGHT = "\ueb70"
    TRIANGLE_UP = "\ueb71"
    TWITTER = "\ueb72"
    UNFOLD = "\ueb73"
    UNLOCK = "\ueb74"
    UNMUTE = "\ueb75"
    UNVERIFIED = "\ueb76"
    VERIFIED = "\ueb77"
    VERSIONS = "\ueb78"
    VM_ACTIVE = "\ueb79"
    VM_OUTLINE = "\ueb7a"
    VM_RUNNING = "\ueb7b"
    WATCH = "\ueb7c"
    WHITESPACE = "\ueb7d"
    WHOLE_WORD = "\ueb7e"
    WINDOW = "\ueb7f"
    WORD_WRAP = "\ueb80"
    ZOOM_IN = "\ueb81"
    ZOOM_OUT = "\ueb82"
    LIST_FILTER = "\ueb83"
    LIST_FLAT = "\ueb84"
    LIST_SELECTION = "\ueb85"
    SELECTION = "\ueb85"
    LIST_TREE = "\ueb86"
    DEBUG_BREAKPOINT_FUNCTION_UNVERIFIED = "\ueb87"
    DEBUG_BREAKPOINT_FUNCTION = "\ueb88"
    DEBUG_BREAKPOINT_FUNCTION_DISABLED = "\ueb88"
    DEBUG_STACKFRAME_ACTIVE = "\ueb89"
    CIRCLE_SMALL_FILLED = "\ueb8a"
    DEBUG_STACKFRAME_DOT = "\ueb8a"
    TERMINAL_DECORATION_MARK = "\ueb8a"
    DEBUG_STACKFRAME = "\ueb8b"
    DEBUG_STACKFRAME_FOCUSED = "\ueb8b"
    DEBUG_BREAKPOINT_UNSUPPORTED = "\ueb8c"
    SYMBOL_STRING = "\ueb8d"
    DEBUG_REVERSE_CONTINUE = "\ueb8e"
    DEBUG_STEP_BACK = "\ueb8f"
    DEBUG_RESTART_FRAME = "\ueb90"
    DEBUG_ALT = "\ueb91"
    CALL_INCOMING = "\ueb92"
    CALL_OUTGOING = "\ueb93"
    MENU = "\ueb94"
    EXPAND_ALL = "\ueb95"
    FEEDBACK = "\ueb96"
    GIT_PULL_REQUEST_REVIEWER = "\ueb96"
    GROUP_BY_REF_TYPE = "\ueb97"
    UNGROUP_BY_REF_TYPE = "\ueb98"
    ACCOUNT = "\ueb99"
    GIT_PULL_REQUEST_ASSIGNEE = "\ueb99"
    BELL_DOT = "\ueb9a"
    DEBUG_CONSOLE = "\ueb9b"
    LIBRARY = "\ueb9c"
    OUTPUT = "\ueb9d"
    RUN_ALL = "\ueb9e"
    SYNC_IGNORED = "\ueb9f"
    PINNED = "\ueba0"
    GITHUB_INVERTED = "\ueba1"
    SERVER_PROCESS = "\ueba2"
    SERVER_ENVIRONMENT = "\ueba3"
    PASS = "\ueba4"
    ISSUE_CLOSED = "\ueba4"
    STOP_CIRCLE = "\ueba5"
    PLAY_CIRCLE = "\ueba6"
    RECORD = "\ueba7"
    DEBUG_ALT_SMALL = "\ueba8"
    VM_CONNECT = "\ueba9"
    CLOUD = "\uebaa"
    MERGE = "\uebab"
    EXPORT = "\uebac"
    GRAPH_LEFT = "\uebad"
    MAGNET = "\uebae"
    NOTEBOOK = "\uebaf"
    REDO = "\uebb0"
    CHECK_ALL = "\uebb1"
    PINNED_DIRTY = "\uebb2"
    PASS_FILLED = "\uebb3"
    CIRCLE_LARGE_FILLED = "\uebb4"
    CIRCLE_LARGE = "\uebb5"
    CIRCLE_LARGE_OUTLINE = "\uebb5"
    COMBINE = "\uebb6"
    GATHER = "\uebb6"
    TABLE = "\uebb7"
    VARIABLE_GROUP = "\uebb8"
    TYPE_HIERARCHY = "\uebb9"
    TYPE_HIERARCHY_SUB = "\uebba"
    TYPE_HIERARCHY_SUPER = "\uebbb"
    GIT_PULL_REQUEST_CREATE = "\uebbc"
    RUN_ABOVE = "\uebbd"
    RUN_BELOW = "\uebbe"
    NOTEBOOK_TEMPLATE = "\uebbf"
    DEBUG_RERUN = "\uebc0"
    WORKSPACE_TRUSTED = "\uebc1"
    WORKSPACE_UNTRUSTED = "\uebc2"
    WORKSPACE_UNKNOWN = "\uebc3"
    TERMINAL_CMD = "\uebc4"
    TERMINAL_DEBIAN = "\uebc5"
    TERMINAL_LINUX = "\uebc6"
    TERMINAL_POWERSHELL = "\uebc7"
    TERMINAL_TMUX = "\uebc8"
    TERMINAL_UBUNTU = "\uebc9"
    TERMINAL_BASH = "\uebca"
    ARROW_SWAP = "\uebcb"
    COPY = "\uebcc"
    PERSON_ADD = "\uebcd"
    FILTER_FILLED = "\uebce"
    WAND = "\uebcf"
    DEBUG_LINE_BY_LINE = "\uebd0"
    INSPECT = "\uebd1"
    LAYERS = "\uebd2"
    LAYERS_DOT = "\uebd3"
    LAYERS_ACTIVE = "\uebd4"
    COMPASS = "\uebd5"
    COMPASS_DOT = "\uebd6"
    COMPASS_ACTIVE = "\uebd7"
    AZURE = "\uebd8"
    ISSUE_DRAFT = "\uebd9"
    GIT_PULL_REQUEST_CLOSED = "\uebda"
    GIT_PULL_REQUEST_DRAFT = "\uebdb"
    DEBUG_ALL = "\uebdc"
    DEBUG_COVERAGE = "\uebdd"
    RUN_ERRORS = "\uebde"
    FOLDER_LIBRARY = "\uebdf"
    DEBUG_CONTINUE_SMALL = "\uebe0"
    BEAKER_STOP = "\uebe1"
    GRAPH_LINE = "\uebe2"
    GRAPH_SCATTER = "\uebe3"
    PIE_CHART = "\uebe4"
    BRACKET = "\ueb0f"
    BRACKET_DOT = "\uebe5"
    BRACKET_ERROR = "\uebe6"
    LOCK_SMALL = "\uebe7"
    AZURE_DEVOPS = "\uebe8"
    VERIFIED_FILLED = "\uebe9"
    NEWLINE = "\uebea"
    LAYOUT = "\uebeb"
    LAYOUT_ACTIVITYBAR_LEFT = "\uebec"
    LAYOUT_ACTIVITYBAR_RIGHT = "\uebed"
    LAYOUT_PANEL_LEFT = "\uebee"
    LAYOUT_PANEL_CENTER = "\uebef"
    LAYOUT_PANEL_JUSTIFY = "\uebf0"
    LAYOUT_PANEL_RIGHT = "\uebf1"
    LAYOUT_PANEL = "\uebf2"
    LAYOUT_SIDEBAR_LEFT = "\uebf3"
    LAYOUT_SIDEBAR_RIGHT = "\uebf4"
    LAYOUT_STATUSBAR = "\uebf5"
    LAYOUT_MENUBAR = "\uebf6"
    LAYOUT_CENTERED = "\uebf7"
    TARGET = "\uebf8"
    INDENT = "\uebf9"
    RECORD_SMALL = "\uebfa"
    ERROR_SMALL = "\uebfb"
    TERMINAL_DECORATION_ERROR = "\uebfb"
    ARROW_CIRCLE_DOWN = "\uebfc"
    ARROW_CIRCLE_LEFT = "\uebfd"
    ARROW_CIRCLE_RIGHT = "\uebfe"
    ARROW_CIRCLE_UP = "\uebff"
    LAYOUT_SIDEBAR_RIGHT_OFF = "\uec00"
    LAYOUT_PANEL_OFF = "\uec01"
    LAYOUT_SIDEBAR_LEFT_OFF = "\uec02"
    BLANK = "\uec03"
    HEART_FILLED = "\uec04"
    MAP = "\uec05"
    MAP_HORIZONTAL = "\uec05"
    FOLD_HORIZONTAL = "\uec05"
    MAP_FILLED = "\uec06"
    MAP_HORIZONTAL_FILLED = "\uec06"
    FOLD_HORIZONTAL_FILLED = "\uec06"
    CIRCLE_SMALL = "\uec07"
    BELL_SLASH = "\uec08"
    BELL_SLASH_DOT = "\uec09"
    COMMENT_UNRESOLVED = "\uec0a"
    GIT_PULL_REQUEST_GO_TO_CHANGES = "\uec0b"
    GIT_PULL_REQUEST_NEW_CHANGES = "\uec0c"
    SEARCH_FUZZY = "\uec0d"
    COMMENT_DRAFT = "\uec0e"
    SEND = "\uec0f"
    SPARKLE = "\uec10"
    INSERT = "\uec11"
    MIC = "\uec12"
    THUMBSDOWN_FILLED = "\uec13"
    THUMBSUP_FILLED = "\uec14"
    COFFEE = "\uec15"
    SNAKE = "\uec16"
    GAME = "\uec17"
    VR = "\uec18"
    CHIP = "\uec19"
    PIANO = "\uec1a"
    MUSIC = "\uec1b"
    MIC_FILLED = "\uec1c"
    REPO_FETCH = "\uec1d"
    COPILOT = "\uec1e"
    LIGHTBULB_SPARKLE = "\uec1f"
    ROBOT = "\uec20"
    SPARKLE_FILLED = "\uec21"
    DIFF_SINGLE = "\uec22"
    DIFF_MULTIPLE = "\uec23"
    SURROUND_WITH = "\uec24"
    SHARE = "\uec25"
    GIT_STASH = "\uec26"
    GIT_STASH_APPLY = "\uec27"
    GIT_STASH_POP = "\uec28"
    VSCODE = "\uec29"
    VSCODE_INSIDERS = "\uec2a"
    CODE_OSS = "\uec2b"
    RUN_COVERAGE = "\uec2c"
    RUN_ALL_COVERAGE = "\uec2d"
    COVERAGE = "\uec2e"
    GITHUB_PROJECT = "\uec2f"
    MAP_VERTICAL = "\uec30"
    FOLD_VERTICAL = "\uec30"
    MAP_VERTICAL_FILLED = "\uec31"
    FOLD_VERTICAL_FILLED = "\uec31"
    GO_TO_SEARCH = "\uec32"
    PERCENTAGE = "\uec33"
    SORT_PERCENTAGE = "\uec33"
    ATTACH = "\uec34"



"""Data classes for the language server client."""

from __future__ import annotations

import dataclasses
from typing import List, Optional

# Requests


@dataclasses.dataclass
class CompletionRequest:
    id: int
    cursor: str


@dataclasses.dataclass
class JumpRequest:
    file_path: str
    location: str


# Responses


@dataclasses.dataclass
class Diagnostic:
    start: str
    end: str
    message: str
    severity: int = 0
    tag: Optional[str] = None

    def __repr__(self) -> str:
        return self.start


@dataclasses.dataclass
class Diagnostics:
    underline_list: List[Diagnostic]


@dataclasses.dataclass
class Completion:
    kind: int
    display_text: str
    replace_start: str
    replace_end: str
    replace_text: str
    filter_text: str
    documentation: str

    def __repr__(self) -> str:
        return self.replace_text


@dataclasses.dataclass
class Completions:
    id: int
    completions: List[Completion]


@dataclasses.dataclass
class HoverResponse:
    location: str
    text: Optional[list[str, str]]
    docs: Optional[str] = None

    def __repr__(self) -> str:
        return self.text


@dataclasses.dataclass
class JumpLocationRange:
    file_path: str
    start: str
    end: str

    def __repr__(self) -> str:
        return self.file_path


@dataclasses.dataclass
class Jump:
    pos: str
    locations: List[JumpLocationRange]


@dataclasses.dataclass
class TextEdit:
    start: str
    end: str
    new_text: str

    def __repr__(self) -> str:
        return self.file_path


@dataclasses.dataclass
class WorkspaceEdit:
    file_path: str
    edits: list[TextEdit]

    def __repr__(self) -> str:
        return self.file_path


@dataclasses.dataclass
class WorkspaceEdits:
    edits: List[WorkspaceEdit]



from __future__ import annotations

import tkinter as tk
import typing
from itertools import chain

from ..ui import Frame

if typing.TYPE_CHECKING:
    from . import Palette


class SearchBar(Frame):
    """Search Bar

    It is used to filter actions based on the search term entered by the user.
    The search bar can be used to search for actions, files, or take input from the user.
    """

    def __init__(self, master: Palette, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.term = ""

        self.text_variable = tk.StringVar()
        self.text_variable.trace_add("write", self.filter)

        # frame = Frame(self, bg=self.base.theme.border)
        # frame.pack(fill=tk.BOTH, padx=2, pady=2)

        self.search_bar = tk.Entry(
            self,
            font=self.base.settings.font,
            width=self.master.width,
            relief=tk.FLAT,
            textvariable=self.text_variable,
            bg=self.base.theme.border,
            fg=self.base.theme.biscuit,
        )

        self.search_bar.grid(sticky=tk.EW, padx=5, pady=5)
        self.configure_bindings()

    def configure_bindings(self) -> None:
        self.search_bar.bind("<Return>", self.master.search_bar_enter)

        self.search_bar.bind("<Down>", lambda e: self.master.select(1))
        self.search_bar.bind("<Up>", lambda e: self.master.select(-1))

    def clear(self) -> None:
        self.text_variable.set("")

    def focus(self) -> None:
        self.search_bar.focus()

    def add_prefix(self, prefix: str = None) -> None:
        """Adds a prefix to the search bar text

        Args:
            prefix (str, optional): The prefix to add. Defaults to None."""

        self.prefix = prefix
        self.text_variable.set(prefix + " " if prefix else "")
        self.search_bar.icursor(tk.END)

    def set_search_term(self, term: str) -> None:
        """Sets the search term

        Args:
            term (str): The search term to set"""

        self.search_bar.focus_set()
        self.text_variable.set(self.search_bar.get() + term)
        self.search_bar.select_range(len(self.prefix), tk.END)
        self.search_bar.icursor(tk.END)

    def get_search_term(self) -> str:
        return self.search_bar.get().lower()

    def filter(self, *args) -> None:
        """Filters the actions based on the search term

        The search term is used to filter the actions based on the following criteria:
        - Exact match
        - Starts with
        - Includes

        If no actions are found, the no results message is displayed."""

        term = self.get_search_term()

        prefix_found = False
        for actionset in self.master.actionsets:
            actionset = actionset()
            if term.startswith(actionset.prefix):
                self.master.pick_actionset(actionset)
                term = term[len(actionset.prefix) :].strip()

                prefix_found = True
                break

        self.term = term
        if not prefix_found:
            self.master.pick_file_search(term)

        exact, starts, includes = [], [], []
        temp = term.lower()
        for i in self.master.active_set:
            if not i or not i[0]:
                continue

            item = i[0].lower()
            if item == temp:
                exact.append(i)
            elif item.startswith(temp):
                starts.append(i)
            elif temp in item:
                includes.append(i)

        new = list(chain(actionset.get_pinned(term), exact, starts, includes))
        self.master.reset_start_index()
        if any(new):
            self.master.show_items(new)
        else:
            self.master.show_no_results()



from ..language.languages import Languages
from .autocomplete import AutoComplete
from .breadcrumbs import BreadCrumbs, PathView
from .comment_prefix import get_comment_prefix, register_comment_prefix
from .diagnostic import Diagnostic
from .editor import Editor
from .editorbase import BaseEditor
from .findreplace import FindReplace
from .hover import Hover
from .html import HTMLEditor
from .image import ImageViewer
from .markdown import MDEditor
from .misc import Welcome
from .peek import Peek
from .rename import Rename
from .text import Text, TextEditor
from .types import EditorTypes



from __future__ import annotations

import os
import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton

if typing.TYPE_CHECKING:
    from .editorsbar import EditorsBar


class TabScroll(Frame):
    """Tab scroll widget

    Tab scroll widget is used to scroll the tabs in the editors bar.
    """

    def __init__(self, master: EditorsBar, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: EditorsBar = master
        self.config(bg=self.base.theme.border)

        self.left = IconButton(
            self,
            Icons.ARROW_LEFT,
            event=self.scroll_right,
            iconsize=12,
            padx=7,
            hfg_only=True,
        )
        self.left.config(**self.base.theme.layout.content.editors.bar)

        self.right = IconButton(
            self,
            Icons.ARROW_RIGHT,
            event=self.scroll_left,
            iconsize=12,
            padx=7,
            hfg_only=True,
        )
        self.right.config(**self.base.theme.layout.content.editors.bar)

        self.left.config(state=tk.DISABLED)
        self.right.config(state=tk.DISABLED)

        self.left.pack(side=tk.LEFT, fill=tk.BOTH)
        self.right.pack(side=tk.RIGHT, fill=tk.BOTH, padx=(0, 1))

    # TODO: Implement scroll_left and scroll_right methods
    def scroll_left(self) -> None: ...
    def scroll_right(self) -> None: ...



import tkinter as tk

import pyperclip

from biscuit.common.ui import Tree
from biscuit.views.debug.menu import VariablesContextMenu

from ..sidebar_item import SideBarViewItem


class Variables(SideBarViewItem):
    """A view that displays the local variables of the debugger.
    Variables are populated when the debugger is running only."""

    def __init__(self, master, *args, **kwargs) -> None:
        self.title = "Variables"
        self.__actions__ = ()
        super().__init__(master, itembar=True, *args, **kwargs)

        self.tree = Tree(
            self.content,
            style="mono.Treeview",
            columns=("key", "value"),
            *args,
            **kwargs,
        )
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)

        self.ctxmenu = VariablesContextMenu(self)

        self.tree.bind("<Button-3>", self.right_click)
        self.tree.tree.tag_configure("bold", font=self.base.settings.uifont_bold)

    def right_click(self, e: tk.Event) -> None:
        if item := self.tree.identify_row(e.y):
            self.tree.selection_set(item)
            self.tree.focus(item)
            self.ctxmenu.show(e)

    def show(self, section_name: str = "", d: dict = {}, open_: bool = False):
        """Show the variables in the tree.

        Args:
            d (dict): mapping of variable names to values
            section (str): section name for grouping variables
        """
        section_node = self.tree.insert(
            "", "end", text=section_name, open=open_, tags=("bold",)
        )
        self._insert_items(d, parent=section_node)

    def _insert_items(self, d: dict, parent: str):
        for key, value in d.items():
            if isinstance(value, dict):
                node = self.tree.insert(
                    parent, "end", text=key, values=(key, str(value))
                )
                self._insert_items(value, parent=node)
            else:
                self.tree.insert(
                    parent,
                    "end",
                    text=f"{key} = {value}",
                    values=(key, str(value)),
                )

    def clear(self):
        """Clear the local variables."""

        try:
            self.tree.delete(*self.tree.get_children())
        except tk.TclError as e:
            print("Debugger variables panel", e)

    def copy_value(self):
        """Copy the value of the selected variable to the clipboard."""
        pyperclip.copy(self.tree.set(self.tree.focus(), "value"))

    def copy_expression(self):
        """Copy the expression of the selected variable to the clipboard."""
        pyperclip.copy(self.tree.item(self.tree.focus(), "text"))

    def copy_name(self):
        """Copy the name of the selected variable to the clipboard."""
        pyperclip.copy(self.tree.set(self.tree.focus(), "key"))

    def set_value(self):
        """Set the value of the selected variable."""

        try:
            self.base.palette.show(
                "debugger.set_local", default=self.tree.set(self.tree.focus(), "value")
            )
        except Exception as e:
            print("Debugger variables:", e)

    def set_variable_callback(self, value: str):
        """Callback function for setting the value of the selected variable."""

        self.base.debugger_manager.latest.set_variable(
            self.tree.set(self.tree.focus(), "key"), value
        )



from biscuit.common.ui import Tree


class PeekTree(Tree):
    def __init__(self, *args, **kwargs):
        super().__init__(columns=("path", "start"), *args, **kwargs)
        self.bind("<<TreeviewSelect>>", self.singleclick)



from biscuit.common import Menu


class ActionMenu(Menu):
    """Action menu for the activity bar

    Action menus are used to display a menu in the activity bar"""

    def get_coords(self, *e) -> list[int, int]:
        return (
            self.master.winfo_rootx() + self.master.winfo_width(),
            self.master.winfo_rooty()
            + self.master.winfo_height()
            - self.winfo_height(),
        )



#https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setwindowpos
#https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-showwindow

import tkinter as tk
from ctypes import windll
from tkinter import ttk


def set_appwindow():
    global hasstyle
    GWL_EXSTYLE=-20
    WS_EX_APPWINDOW=0x00040000
    WS_EX_TOOLWINDOW=0x00000080
    if not hasstyle:
        hwnd = windll.user32.GetParent(root.winfo_id())
        style = windll.user32.GetWindowLongW(hwnd, GWL_EXSTYLE)
        style = style & ~WS_EX_TOOLWINDOW
        style = style | WS_EX_APPWINDOW
        res = windll.user32.SetWindowLongW(hwnd, GWL_EXSTYLE, style)
        root.withdraw()
        root.after(10, root.wm_deiconify)
        hasstyle=True

def get_pos(event: tk.Event):
    global xwin
    global ywin
    xwin = event.x
    ywin = event.y

def move_window(event: tk.Event):
    global previousPosition
    root.geometry(f'+{event.x_root - xwin}+{event.y_root - ywin}')
    previousPosition = [root.winfo_x(), root.winfo_y()]

def move_window_bindings(*args, status=True):
    if status is True:
        title_bar.bind("<B1-Motion>", move_window)
        title_bar.bind("<Button-1>", get_pos)
        title_name.bind("<B1-Motion>", move_window)
        title_name.bind("<Button-1>", get_pos)
    else:
        title_bar.unbind("<B1-Motion>")
        title_bar.unbind("<Button-1>")
        title_name.unbind("<B1-Motion>")
        title_name.unbind("<Button-1>")

def quit():
    root.destroy()

#reference: https://programtalk.com/python-examples/ctypes.windll.user32.ShowWindow/
def minimize(hide=False):
    hwnd = windll.user32.GetParent(root.winfo_id())
    windll.user32.ShowWindow(hwnd, 0 if hide else 6)

def maximizeToggle():
    global maximized
    if not maximized:
        #maximize current window
        maximize_btn.config(text="❐")
        hwnd = windll.user32.GetParent(root.winfo_id())
        SWP_SHOWWINDOW = 0x40
        windll.user32.SetWindowPos(hwnd, 0, 0, 0, int(root.winfo_screenwidth()), int(root.winfo_screenheight()-48),SWP_SHOWWINDOW)
        maximized = True
        move_window_bindings(status=False)
    else:
        #restore down window
        maximize_btn.config(text="🗖")
        hwnd = windll.user32.GetParent(root.winfo_id())
        SWP_SHOWWINDOW = 0x40
        windll.user32.SetWindowPos(hwnd, 0, previousPosition[0], previousPosition[1], int(root.minsize()[0]), int(root.minsize()[1]),SWP_SHOWWINDOW)
        maximized = False
        move_window_bindings(status=True)


#---------------------------------
root = tk.Tk()
root.overrideredirect(True)

#window details
maximized = False
back_ground = "#2c2c2c"
dimension = (300, 300)
#------------------------------

if len(dimension) == 0:
    #default window dimension
    x = (root.winfo_screenwidth()/2)-(350/2)
    y = (root.winfo_screenheight()/2)-(250)
    root.geometry(f'350x150+{int(x)}+{int(y)}')
    root.minsize(350, 150)
    dimension = (350, 150)
    previousPosition = [int(x), int(y)]

else:
    x = (root.winfo_screenwidth()/2)-(dimension[0]/2)
    y = (root.winfo_screenheight()/2)-250
    root.geometry(f'{dimension[0]}x{dimension[1]}+{int(x)}+{int(y)}')
    root.minsize(dimension[0], dimension[1])
    previousPosition = [int(x), int(y)]




#title bar
title_bar = tk.Frame(root, bg=back_ground, bd=1,
                     highlightcolor=back_ground, 
                     highlightthickness=0)

#window title
title_window = "Untitled window"
title_name = tk.Label(title_bar, text=title_window, 
                     font="Arial 12", bg=back_ground, fg="white")

#minimize btn
minimize_btn = tk.Button(title_bar, text='🗕', bg=back_ground, padx=5, pady=2, 
                         bd=0, font="bold", fg='white', width=2,
                         activebackground="red",
                         activeforeground="white", 
                         highlightthickness=0, 
                         command=minimize)

#maximize btn
maximize_btn = tk.Button(title_bar, text='🗖', bg=back_ground, padx=5, pady=2, 
                         bd=0, font="bold", fg='white', width=2,
                         activebackground="red",
                         activeforeground="white", 
                         highlightthickness=0, 
                         command=maximizeToggle)

#close btn
close_button = tk.Button(title_bar, text='🗙', bg=back_ground, padx=5, pady=2, 
                         bd=0, font="bold", fg='white', width=2,
                         activebackground="red",
                         activeforeground="white", 
                         highlightthickness=0, 
                         command= quit)

#hover effect
minimize_btn.bind('<Enter>', lambda x: minimize_btn.configure(bg='#777777'))
minimize_btn.bind('<Leave>', lambda x: minimize_btn.configure(bg=back_ground))
maximize_btn.bind('<Enter>', lambda x: maximize_btn.configure(bg='#777777'))
maximize_btn.bind('<Leave>', lambda x: maximize_btn.configure(bg=back_ground))
close_button.bind('<Enter>', lambda x: close_button.configure(bg='red'))
close_button.bind('<Leave>',lambda x: close_button.configure(bg=back_ground))


#main area of the window
window = tk.Frame(root, bg="white", highlightthickness=1, highlightbackground=back_ground)

txt = tk.Label(window, bg='white', text="Prototype window").pack(anchor="center")

# pack the widgets
title_bar.pack(fill='x', side=tk.TOP)
title_name.pack(side='left', padx=5)
close_button.pack(side='right')
maximize_btn.pack(side=tk.RIGHT)
minimize_btn.pack(side=tk.RIGHT)
window.pack(fill='both', expand=True, side=tk.TOP)
move_window_bindings(status=True)


#ctype
hasstyle = False
root.update_idletasks()

root.mainloop()



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from biscuit import App
    from biscuit.editor import BaseEditor


class ExecutionManager:
    """Execution Manager

    The ExecutionManager class is used to run commands in the terminal.
    - provides a method to register a command for a language
    - generate run command for editor based on language
    - run a command in the integrated terminal / external terminal
    """

    def __init__(self, base: App):
        self.base = base

        self.commands = {"Python": "python"}

    def register_command(self, language: str, command: str) -> None:
        self.commands[language] = command

    def run_command(self, editor: BaseEditor, external=False) -> None:
        if external:
            self.base.terminalmanager.run_external_console(
                f"{editor.run_command_value} &"
            )
            return

        self.base.terminalmanager.run_command(editor.run_command_value)
        self.base.terminalmanager.active_terminal.enter()

    def get_command(self, editor: BaseEditor) -> str:
        if c := self.commands.get(editor.language, None):
            return f"{c} {editor.path}"



import tkinter as tk
from typing import Callable

from biscuit.common.icons import Icons

from ..ui import Frame, IconButton, Label, Toplevel
from .notification import Notification


class Notifications(Toplevel):
    """
    Floating notifications, shown on bottom right corner

    Attributes:
        active (bool): whether shown or not
        count (int): number of notifications
        notifications (list[Notification]): list of active notifications
        title (Label): title bar label
        xoffset (int): x offset
        yoffset (int): y offset

    Methods:
        info: create info notification
        warning: create warning notification
        error: create error notification
        show: show notification
        hide: hide notification
        clear: clear all notifications
        delete: delete notification
    """

    def __init__(self, base) -> None:
        super().__init__(base)
        self.config(bg=self.base.theme.border, padx=1, pady=1)
        self.active = False
        self.overrideredirect(True)

        self.minsize(
            width=round(300 * self.base.scale), height=round(15 * self.base.scale)
        )
        self.minsize(
            width=round(300 * self.base.scale), height=round(15 * self.base.scale)
        )
        self.withdraw()
        self.xoffset = 10 * self.base.scale
        self.yoffset = 40 * self.base.scale

        self.latest: Notification = None
        self.count = 0
        self.toggle_button: IconButton = None
        self.notifications: list[Notification] = []

        topbar = Frame(self, **self.base.theme.notifications)
        topbar.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        self.title = Label(
            topbar,
            text="NOTIFICATIONS",
            anchor=tk.W,
            font=self.base.settings.uifont,
            **self.base.theme.notifications.title,
        )
        self.title.pack(side=tk.LEFT, padx=10, fill=tk.BOTH, expand=1)

        close_button = IconButton(topbar, Icons.CLOSE, self.hide)
        close_button.config(**self.base.theme.notifications.title)
        close_button.pack(side=tk.RIGHT, fill=tk.BOTH, pady=(0, 1))

        # self.base.bind("<FocusIn>", lambda *_: self.lift, add=True)
        # self.base.bind("<Configure>", self.refresh_position, add=True)

        self.bind("<FocusIn>", lambda *_: self.lift(), add=True)
        self.bind("<FocusOut>", self.hide, add=True)

        self.lift()
        self.withdraw()

    def info(
        self, text: str, actions: list[tuple[str, Callable[[None], None]]] = None
    ) -> Notification:
        """Create an info notification

        Args:
            text (str): notification text"""

        instance = Notification(
            self, Icons.INFO, text=text, fg=self.base.theme.biscuit, actions=actions
        )
        instance.pack(side=tk.TOP, fill=tk.BOTH, expand=1, pady=(0, 1))
        self.count += 1
        self.show()
        self.update_idletasks()
        self.refresh_position()

        self.notifications.append(instance)
        self.latest = instance
        return instance

    def warning(
        self, text: str, actions: list[tuple[str, Callable[[None], None]]] = None
    ) -> Notification:
        """Create a warning notification

        Args:
            text (str): notification text"""

        instance = Notification(
            self, Icons.WARNING, text=text, fg="yellow", actions=actions
        )
        instance.pack(side=tk.TOP, fill=tk.BOTH, expand=1, pady=(0, 1))
        self.count += 1
        self.show()
        self.update_idletasks()
        self.refresh_position()

        self.notifications.append(instance)
        self.latest = instance
        return instance

    def error(
        self, text: str, actions: list[tuple[str, Callable[[None], None]]] = None
    ) -> Notification:
        """Create an error notification

        Args:
            text (str): notification text"""

        instance = Notification(self, Icons.ERROR, text=text, fg="red", actions=actions)
        instance.pack(side=tk.TOP, fill=tk.BOTH, expand=1, pady=(0, 1))
        self.count += 1
        self.show()
        self.update_idletasks()
        self.refresh_position()

        self.notifications.append(instance)
        self.latest = instance
        return instance

    def notify(
        self,
        text: str,
        kind: int,
        actions: list[tuple[str, Callable[[None], None]]] = None,
    ) -> Notification:
        """Create a notification based on kind
        1: info
        2: warning
        3: error

        Args:
            text (str): notification text
            kind (int): notification kind
        """
        match kind:
            case 1:
                self.error(text, actions)
            case 2:
                self.warning(text, actions)
            case _:
                self.info(text, actions)

    def set_button(self, button: IconButton) -> None:
        """Set button to show notifications

        Args:
            button (IconButton): button to show notifications"""

        self.toggle_button = button

    def refresh_position(self, *_) -> None:
        """Follow root window position"""

        if not self.active:
            return

        try:
            # x = (
            #     self.base.winfo_x()
            #     + self.base.winfo_width()
            #     - self.winfo_width()
            #     - self.xoffset
            # )
            # y = (
            #     self.base.winfo_y()
            #     + self.base.winfo_height()
            #     - self.winfo_height()
            #     - self.yoffset
            # )

            x = (
                self.toggle_button.winfo_rootx()
                + self.toggle_button.winfo_width()
                - self.winfo_width()
            )
            y = self.toggle_button.winfo_rooty() + self.toggle_button.winfo_height() - 2

            self.geometry(f"+{int(x)}+{int(y)}")
        except tk.TclError:
            # root window is destroyed
            pass

    def toggle(self, *_) -> None:
        """Toggle notification visibility"""

        if self.active:
            self.hide()
        else:
            self.show()

    def show(self, *_) -> None:
        """Toggle notification visibility
        Also updates title based on count."""

        if self.count:
            self.title.config(text=f"NOTIFICATIONS ({self.count})")
        else:
            self.title.config(text="NO NEW NOTIFICATIONS")

        self.active = True
        self.deiconify()
        self.refresh_position()
        self.focus_force()
        self.base.menubar.update_notifications()

    def hide(self, *_) -> None:
        """Hide notifications"""

        self.active = False
        self.withdraw()

    def clear(self, *_) -> None:
        """Clear all notifications"""

        for i in self.notifications:
            i.delete()
        self.count = 0

    def delete_notification(self, notification: Notification) -> None:
        """Delete a notification

        Args:
            notification (Notification): notification to delete"""

        self.notifications.remove(notification)
        notification.destroy()
        self.count -= 1

        self.update_idletasks()
        self.refresh_position()

        if not self.count:
            self.hide()

        self.base.menubar.update_notifications()



from __future__ import annotations

import tkinter as tk
import typing

if typing.TYPE_CHECKING:
    from .text import Text
    from .editor import TextEditor

from biscuit.common.ui import Canvas, Menubutton


class LineNumbers(Canvas):
    """Line Numbers widget

    This widget is used to display line numbers for the text widget.
    It also provides the functionality to toggle breakpoints on the lines.
    """

    def __init__(self, master: TextEditor, text: Text = None, *args, **kwargs) -> None:
        """Line Numbers widget

        Args:
            master: The parent widget
            text (Text, optional): The text widget to attach. Defaults to None.
        """

        super().__init__(master, *args, **kwargs)
        self.master: TextEditor = master
        self.config(
            width=65, bd=0, highlightthickness=0, **self.base.theme.editors.linenumbers
        )
        self.text = text
        self.font = self.base.settings.font

        self.bg, self.fg, _, self.hfg = (
            self.base.theme.editors.linenumbers.number.values()
        )
        self.bp_hover_color, _, self.bp_enabled_color, _ = (
            self.base.theme.editors.linenumbers.breakpoint.values()
        )
        self.breakpoints: set[int] = set()

    def attach(self, text):
        self.text = text

    def mark_line(self, line):
        dline = self.text.dlineinfo(line)

        if not dline:
            return

        y = dline[1]
        btn = Menubutton(
            self,
            text=">",
            font=self.base.settings.font,
            cursor="hand2",
            borderwidth=0,
            width=2,
            height=1,
            pady=0,
            padx=0,
            relief=tk.FLAT,
            **self.base.theme.linenumbers,
        )
        self.create_window(70, y - 2, anchor=tk.NE, window=btn)

    def set_bar_width(self, width):
        self.configure(width=width)

    def toggle_breakpoint(self, line):
        if line in self.breakpoints:
            self.breakpoints.remove(line)
        else:
            self.breakpoints.add(line)
        self.redraw()
        self.master.update_breakpoints(self.breakpoints)

    def redraw(self, *_):
        self.delete(tk.ALL)

        if not self.text:
            return

        current_line = int(self.text.index(tk.INSERT).split('.')[0]) 

        i = self.text.index("@0,0")
        while True:
            dline = self.text.dlineinfo(i)
            if dline is None:
                break

            y = dline[1]
            linenum = int(float(i))
            curline = self.text.dlineinfo(tk.INSERT)
            cur_y = curline[1] if curline else None

            # TODO: optional: skipping comments
            #             # check if line is commented
            # if self.text.highlighter and any((tag[1].startswith("Token.Comment") for tag in self.text.dump(i, i + " lineend", tag=True))):
            # if self.text.get(f"{i} linestart", f"{i} lineend").strip().startswith("#"):
            #     self.create_text(40, y, anchor=tk.NE, text="|   ", tag=i, fill=self.base.theme.border)
            #     i = self.text.index(f"{i}+1line")
            #     continue

            # Render breakpoint
            has_breakpoint = linenum in self.breakpoints
            breakpoint_id = self.create_oval(
                5,
                y + 3,
                15,
                y + 13,
                outline="",
                fill=self.bg if not has_breakpoint else self.bp_enabled_color,
            )

            # Bind hover and click events for breakpoint
            self.tag_bind(
                breakpoint_id,
                "<Enter>",
                lambda _, breakpoint_id=breakpoint_id, flag=has_breakpoint: self.on_breakpoint_enter(
                    breakpoint_id, flag
                ),
            )
            self.tag_bind(
                breakpoint_id,
                "<Leave>",
                lambda _, breakpoint_id=breakpoint_id, flag=has_breakpoint: self.on_breakpoint_leave(
                    breakpoint_id, flag
                ),
            )
            self.tag_bind(
                breakpoint_id,
                "<Button-1>",
                lambda _, linenum=linenum: self.toggle_breakpoint(linenum),
            )

            if self.text.relative_line_numbers:
                if linenum != current_line:
                    linenum = abs(linenum - current_line)

            self.create_text(
                40,
                y,
                anchor=tk.NE,
                text=linenum,
                font=self.font,
                tag=i,
                fill=self.hfg if y == cur_y else self.fg,
            )

            self.tag_bind(
                i,
                "<Enter>",
                lambda _, breakpoint_id=breakpoint_id, flag=has_breakpoint: self.on_breakpoint_enter(
                    breakpoint_id, flag
                ),
            )

            self.tag_bind(
                i,
                "<Leave>",
                lambda _, breakpoint_id=breakpoint_id, flag=has_breakpoint: self.on_breakpoint_leave(
                    breakpoint_id, flag
                ),
            )

            i = self.text.index(f"{i}+1line")

    def on_breakpoint_enter(self, id, flag):
        self.itemconfig(id, fill=self.bp_enabled_color if flag else self.bp_hover_color)

    def on_breakpoint_leave(self, id, flag):
        self.itemconfig(id, fill=self.bp_enabled_color if flag else self.bg)



import tkinter as tk

import tarts as lsp

from biscuit.common.ui import Text
from biscuit.language.utils import decode_position

from .kinds import kinds


class Tree(Text):
    """Tree view that displays the outline of the active document."""

    def __init__(self, master):
        super().__init__(master, wrap="none", borderwidth=0, highlightthickness=0)
        self.master = master
        self.base = master.base
        self.config(
            cursor="hand2",
            **self.base.theme.views.sidebar.item.content,
            font=self.base.settings.uifont,
            padx=10,
            pady=10,
            spacing1=0,
            spacing2=0,
            spacing3=0,
        )

        for kind in kinds:
            if kind[1]:
                self.tag_config(kind[0], foreground=kind[1])
            self.tag_config(kind[0], font="codicon 12")
        self.tag_config(
            "branch", foreground=self.base.theme.border, font=("Segoi UI", 15)
        )

        # self.bind("<Button-1>", self.onclick)
        # self.bind("<Enter>", self.hoverin)
        # self.bind("<Leave>", self.hoverout)

    # def hoverin(self, _: tk.Event) -> None:
    #     self.tag_config("branch", foreground=self.base.theme.border)

    # def hoverout(self, _: tk.Event) -> None:
    #     self.tag_config("branch", foreground=self.base.theme.views.sidebar.item.content.background)

    def add_items(self, items: list[lsp.DocumentSymbol], level=0) -> None:
        if not items:
            return

        for item in items:
            if item.kind == lsp.SymbolKind.MODULE:
                continue

            tag = f"{item.name}{item.kind}"
            self.insert(tk.END, "┊" * level, "branch")
            self.insert(tk.END, kinds[item.kind - 1][0], kinds[item.kind - 1][0])
            self.insert(tk.END, f" {item.name}\n", tag)
            self.add_items(item.children, level + 1)
            self.tag_bind(
                tag,
                "<Button-1>",
                lambda e, i=item: self.base.goto_location_in_active_editor(
                    decode_position(i.range.start)
                ),
            )

    # def onclick(self, event: tk.Event) -> None:
    #     self.config(state=tk.NORMAL)
    #     index = self.index(f"@{event.x},{event.y}")
    #     self.tag_add(tk.SEL, f'{index} linestart', f'{index} lineend+1c')
    #     self.config(state=tk.DISABLED)



from __future__ import annotations

import tkinter as tk
import typing

from ..ui import Text

if typing.TYPE_CHECKING:
    from . import Palette


class PaletteItem(Text):
    """Palette Item

    Palette Item is a text widget that represents an action that can be performed by the user.
    It is displayed in the palette and can be selected by the user using the mouse or keyboard.
    Text widget is used to highlight the search term in the item text.
    """

    def __init__(
        self, master: Palette, text: str, command: str, description="", *args, **kwargs
    ) -> None:
        """Initializes the Palette Item

        Args:
            master (Palette): The parent palette instance
            text (str): The text to display in the item
            command (str): The command to execute when the item is selected
            description (str, optional): The description of the item. Defaults to"""

        super().__init__(master, *args, **kwargs)
        self.text = text
        self.description = description
        self.command = command

        self.bg, self.fg, self.hbg, self.hfg = self.base.theme.palette.item.values()
        self.config(
            font=self.base.settings.uifont,
            fg=self.fg,
            bg=self.bg,
            cursor="hand2",
            padx=10,
            pady=3,
            relief=tk.FLAT,
            highlightthickness=0,
            width=30,
            height=1,
        )

        self.tag_config("term", foreground=self.base.theme.biscuit)
        self.tag_config("description", foreground=self.base.theme.primary_foreground)

        self.insert(tk.END, text)
        self.insert(tk.END, f" {description}", "description")
        self.config(state=tk.DISABLED)

        self.bind("<Button-1>", self.on_click)
        self.bind("<Enter>", self.on_hover)
        self.bind("<Leave>", self.off_hover)

        self.selected = False
        self.hovered = False

    def on_click(self, *args) -> None:
        """Executes the command when the item is clicked"""

        term = self.master.searchbar.term

        self.master.hide()
        self.command(term)

    def toggle_selection(self) -> None:
        """Toggles the selection state of the item"""

        if self.selected:
            self.select()
        else:
            self.deselect()

    def mark_term(self, term: str) -> None:
        """Marks the search term in the item text"""

        start_pos = self.text.lower().find(term.lower())
        end_pos = start_pos + len(term)
        self.tag_remove("term", 1.0, tk.END)
        self.tag_add("term", f"1.{start_pos}", f"1.{end_pos}")

    def on_hover(self, *args) -> None:
        if not self.selected:
            self.config(bg=self.hbg)
            self.hovered = True

    def off_hover(self, *args) -> None:
        if not self.selected:
            self.config(bg=self.bg)
            self.hovered = False

    def toggle_selection(self) -> None:
        if self.selected:
            self.select()
        else:
            self.deselect()

    def select(self) -> None:
        self.config(bg=self.hbg, fg=self.hfg)
        self.selected = True

    def deselect(self) -> None:
        self.config(bg=self.bg, fg=self.fg)
        self.selected = False



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common import Icons

from ..helpers import caller_class_name
from ..ui import Frame, Icon, IconButton, IconLabelButton, Label

if typing.TYPE_CHECKING:
    from .manager import Notifications


class Notification(Frame):
    """Notification class

    Holds the notification icon, text and close button"""

    def __init__(
        self,
        master: Notifications,
        icon: str,
        text: str,
        actions: list[tuple[str, typing.Callable[[None], None]]],
        fg: str,
        *args,
        **kwargs,
    ) -> None:
        """Create a notification

        Args:
            master: Parent widget
            icon (str): Icon name
            text (str): Notification text
            fg (str): Foreground color
            actions (list[tuple(str, Callable[[None], None])]): List of actions"""

        super().__init__(master, *args, **kwargs)
        self.master: Notifications = master
        self.config(**self.base.theme.notifications)

        top = Frame(self, **self.base.theme.utils.frame)
        top.pack(fill=tk.BOTH, expand=1)
        bottom = Frame(self, **self.base.theme.utils.frame)
        bottom.pack(fill=tk.BOTH)

        self.icon = Icon(top, icon, padx=5, **self.base.theme.utils.iconbutton)
        self.icon.config(fg=fg)
        self.icon.pack(side=tk.LEFT, fill=tk.BOTH)

        self.label = Label(
            top,
            text=text,
            anchor=tk.W,
            padx=10,
            pady=10,
            font=self.base.settings.uifont,
            **self.base.theme.utils.iconbutton,
        )
        self.label.pack(side=tk.LEFT, expand=1, fill=tk.BOTH, anchor=tk.W)

        close_button = IconButton(top, Icons.CLOSE, self.delete)
        close_button.pack(fill=tk.BOTH)

        self.source_label = Label(
            bottom,
            text=f"Source: {caller_class_name(3)}",
            anchor=tk.W,
            padx=10,
            pady=5,
            font=self.base.settings.uifont,
            **self.base.theme.notifications.source,
        )
        self.source_label.pack(side=tk.LEFT, fill=tk.BOTH)

        if actions:
            self.actions = Frame(bottom, **self.base.theme.notifications)
            self.actions.pack(side=tk.RIGHT, fill=tk.BOTH)

            for text, action in actions:
                btn = IconLabelButton(
                    self.actions, text, callback=action, highlighted=True
                )
                btn.pack(side=tk.LEFT, fill=tk.BOTH, padx=(0, 5), pady=(0, 5))

    def delete(self):
        """Delete the notification"""

        try:
            self.master.delete_notification(self)
        except Exception as e:
            self.base.logger.error(e)



import json
import tkinter as tk
import typing
import webbrowser
from tkinter import ttk

import requests

from biscuit.common import ActionSet
from biscuit.common.ui import Scrollbar
from biscuit.git import IssueViewer

from ..sidebar_item import SideBarViewItem


class Issues(SideBarViewItem):
    """View that displays the open issues in a GitHub repository.

    The Issues view displays the open issues in a GitHub repository.
    - The user can click on an issue to open it in the browser.
    """

    def __init__(self, master, itembar=True, *args, **kwargs) -> None:
        self.title = "Open Issues"
        self.__actions__ = ()
        super().__init__(master, itembar=itembar, *args, **kwargs)

        self.url_template = "https://api.github.com/repos/{}/{}/issues"
        self.url = None
        self.owner = None
        self.repo = None

        self.issues = {}

        self.tree = ttk.Treeview(
            self.content,
            selectmode=tk.BROWSE,
            show="tree",
            displaycolumns="",
            columns=("number"),
        )
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.tree.bind("<Double-Button-1>", self.on_click)

        self.scrollbar = Scrollbar(
            self.content,
            style="TreeScrollbar",
            orient=tk.VERTICAL,
            command=self.tree.yview,
        )
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)
        self.tree.config(yscrollcommand=self.scrollbar.set)

        self.issues_actionset = ActionSet("Search GitHub issues", "issue:", [])
        self.base.palette.register_actionset(lambda: self.issues_actionset)

    def set_url(self, owner: str, repo: str) -> None:
        """Sets the URL for the current repository."""

        self.owner = owner
        self.repo = repo
        self.url = self.url_template.format(owner, repo)

    def on_click(self, *_) -> None:
        """Event handler for treeview item click event."""

        item = self.tree.selection()[0]
        number = self.tree.item(item, "values")[0]
        self.open_editor(number)

    def open_editor(self, number: str) -> None:
        """Opens the issue viewer for the selected issue."""

        self.base.editorsmanager.add_editor(
            IssueViewer(self.base.editorsmanager, self.issues[number])
        )

    def fetch(self) -> typing.List[dict]:
        """Fetches issues from the current repository."""

        response = requests.get(self.url)
        if response.status_code != 200:
            self.base.notifications.error(
                f"Failed to fetch issues from {self.owner}/{self.repo}"
            )
            self.base.logger.error(
                f"Failed to fetch issues from {self.owner}/{self.repo}: {response.reason} {response.text}"
            )
            return

        issues = json.loads(response.text)
        self.issues = {str(issue["number"]): issue for issue in issues}
        if not issues:
            self.tree.insert("", tk.END, text="No open issues")
            return

        self.issues_actionset.update(
            [
                (
                    f"{issue['title']} #{issue['number']}",
                    lambda *_, number=issue["number"]: self.open_editor(number),
                )
                for issue in issues
                if "pull_request" not in issue
            ]
        )
        issues = (
            (f"{issue['title']} #{issue['number']}", issue["number"])
            for issue in issues
            if "pull_request" not in issue
        )

        self.tree.delete(*self.tree.get_children())
        for issue in issues:
            self.tree.insert("", tk.END, text=issue[0], values=(issue[1],))



from biscuit.common.icons import Icons

kinds = (
    (Icons.SYMBOL_FILE, None),
    (Icons.SYMBOL_FILE, None),
    (Icons.SYMBOL_NAMESPACE, None),
    (Icons.PACKAGE, None),
    (Icons.SYMBOL_CLASS, "#de9327"),
    (Icons.SYMBOL_METHOD, "#AB7CCF"),
    (Icons.SYMBOL_PROPERTY, "#6eb1ee"),
    (Icons.SYMBOL_FIELD, "#6eb1ee"),
    (Icons.SYMBOL_METHOD, "#6eb1ee"),
    (Icons.SYMBOL_ENUM, None),
    (Icons.SYMBOL_INTERFACE, "#de9327"),
    (Icons.SYMBOL_METHOD, "#AB7CCF"),
    (Icons.SYMBOL_VARIABLE, "#6eb1ee"),
    (Icons.SYMBOL_CONSTANT, None),
    (Icons.SYMBOL_STRING, None),
    (Icons.SYMBOL_NUMERIC, None),
    (Icons.SYMBOL_BOOLEAN, "#de9327"),
    (Icons.SYMBOL_ARRAY, None),
    (Icons.SYMBOL_CLASS, None),
    (Icons.SYMBOL_KEY, None),
    (Icons.SYMBOL_MISC, None),
    (Icons.SYMBOL_ENUM_MEMBER, None),
    (Icons.SYMBOL_STRUCTURE, None),
    (Icons.SYMBOL_EVENT, "#ffd700"),
    (Icons.SYMBOL_OPERATOR, None),
    (Icons.SYMBOL_PARAMETER, "#ffd700"),
)



import random
import tkinter as tk

from biscuit.common.ui import Button, Canvas

from .gamebase import BaseGame


class TicTacToe(BaseGame):
    name = "Tic-Tac-Toe"

    def __init__(self, master, *args, **kwargs):
        super().__init__(master, None, None, *args, **kwargs)

        self.player = self.base.theme.biscuit
        self.computer = self.base.theme.border
        self.border = self.base.theme.border

        self.canvas = Canvas(self, width=300, height=300, **self.base.theme.editors)
        self.canvas.pack(pady=100)

        self.start()
        self.canvas.bind("<Button-1>", self.handle_click)

    def start(self, *_):
        self.canvas.delete("all")
        self.board = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        self.draw_board()

    def draw_board(self):
        self.canvas.delete("all")

        for row in range(1, 3):
            self.canvas.create_line(
                0, row * 100, 300, row * 100, width=2, fill=self.border
            )
        for col in range(1, 3):
            self.canvas.create_line(
                col * 100, 0, col * 100, 300, width=2, fill=self.border
            )

        for row in range(3):
            for col in range(3):
                x = col * 100
                y = row * 100

                if self.board[row][col] == 1:
                    self.canvas.create_line(
                        x + 15, y + 15, x + 85, y + 85, width=10, fill=self.player
                    )
                    self.canvas.create_line(
                        x + 15, y + 85, x + 85, y + 15, width=10, fill=self.player
                    )
                elif self.board[row][col] == 2:
                    self.canvas.create_oval(
                        x + 15, y + 15, x + 85, y + 85, width=10, outline=self.computer
                    )

    def handle_click(self, event: tk.Event):
        row = event.y // 100
        col = event.x // 100

        if self.board[row][col] == 0:
            self.board[row][col] = 1
            self.draw_board()
            self.computer_move()

    def computer_move(self):
        for row in range(3):
            for col in range(3):
                if self.board[row][col] == 0:
                    self.board[row][col] = 2
                    if self.check_winner(2):
                        self.draw_board()
                        self.display_winner(2)
                        return
                    self.board[row][col] = 0

        for row in range(3):
            for col in range(3):
                if self.board[row][col] == 0:
                    self.board[row][col] = 1
                    if self.check_winner(1):
                        self.board[row][col] = 2  # Block the player's winning move
                        self.draw_board()
                        return
                    self.board[row][col] = 0

        # Choose a random move
        available_moves = [
            (row, col)
            for row in range(3)
            for col in range(3)
            if self.board[row][col] == 0
        ]
        if not available_moves:
            return self.display_winner(0)

        row, col = random.choice(available_moves)
        self.board[row][col] = 2
        self.draw_board()

        if not self.check_winner(1) and not any(0 in row for row in self.board):
            self.display_winner(0)

    def check_winner(self, player):
        # Check rows
        for row in range(3):
            if all(self.board[row][col] == player for col in range(3)):
                return True

        # Check columns
        for col in range(3):
            if all(self.board[row][col] == player for row in range(3)):
                return True

        # Check diagonals
        if all(self.board[i][i] == player for i in range(3)) or all(
            self.board[i][2 - i] == player for i in range(3)
        ):
            return True

        return False

    def display_winner(self, winner):
        if not winner:
            message = "It's a tie!"
        else:
            if winner == 1:
                message = "You win!"
            else:
                message = "Torte wins!"

        self.canvas.delete("all")
        self.canvas.create_text(
            150, 125, text=message, font=("Fixedsys", 20), fill=self.player
        )

        restart = Button(self.canvas, "Retry!", self.start)
        restart.config(font=("Fixedsys", 20))
        self.canvas.create_window(150, 175, window=restart)



import tkinter as tk

from biscuit.common.ui import Frame

from .gamebase import BaseGame
from .gameoflife import GameOfLife
from .minesweeper import Minesweeper
from .pong import Pong
from .ttt import TicTacToe
from .whoops import Whoops


class GameManager:
    def __init__(self, base) -> None:
        self.base = base

        self.games = {i.name: i for i in (GameOfLife, Pong, TicTacToe, Minesweeper)}

    def get_games(self) -> list:
        """For palette to generate action sets of games"""

        return [
            (f"Play {i}", lambda _, i=i: self.base.open_game(i))
            for i in self.games.keys()
        ]

    def get_game(self, name) -> str:
        """returns the game class from the name

        Args:
            name (str): name of the game
        """

        return self.games.get(name, Whoops)

    def register_new_game(self, game: BaseGame) -> None:
        """Registers a game to the games dict

        Args:
            game (BaseGame): game to be registered
        """

        try:
            self.games[game.name] = game
        except AttributeError:
            self.games[f"Game {len(self.games) + 1}"] = game


class Game(Frame):
    """Responsible for picking the right game to display and displaying it"""

    def __init__(self, master, name, *args, **kwargs) -> None:
        """Initializes the game frame

        Args:
            master (tk.Tk): master window
            name (str): name of the game"""

        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.filename = name

        self.path = None
        self.exists = False
        self.diff = False
        self.showpath = False

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.content = self.base.game_manager.get_game(name=name)(self)
        self.content.grid(row=0, column=0, sticky=tk.NSEW)

    def focus(self) -> None:
        """Focuses on the game"""

        self.content.focus_get()



import json
import tkinter as tk
import typing
import webbrowser
from tkinter import ttk

import requests

from biscuit.common import ActionSet
from biscuit.common.ui import Scrollbar
from biscuit.git import PRViewer

from ..sidebar_item import SideBarViewItem


class PRs(SideBarViewItem):
    """View that displays the open pull requests in a GitHub repository.

    The PRs view displays the open pull requests in a GitHub repository.
    - The user can click on a PR to open it in the browser.
    """

    def __init__(self, master, itembar=True, *args, **kwargs) -> None:
        self.title = "Pull Requests"
        self.__actions__ = ()
        super().__init__(master, itembar=itembar, *args, **kwargs)

        self.url_template = "https://api.github.com/repos/{}/{}/pulls"
        self.url = None
        self.owner = None
        self.repo = None

        self.prs = {}

        self.tree = ttk.Treeview(
            self.content,
            selectmode=tk.BROWSE,
            show="tree",
            displaycolumns="",
            columns=("number"),
        )
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.tree.bind("<Double-Button-1>", self.on_click)

        self.scrollbar = Scrollbar(
            self.content,
            style="TreeScrollbar",
            orient=tk.VERTICAL,
            command=self.tree.yview,
        )
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)
        self.tree.config(yscrollcommand=self.scrollbar.set)

        self.prs_actionset = ActionSet("Search GitHub pull requests", "pr:", [])
        self.base.palette.register_actionset(lambda: self.prs_actionset)

    def set_url(self, owner: str, repo: str) -> None:
        """Sets the URL for the current repository."""

        self.owner = owner
        self.repo = repo
        self.url = self.url_template.format(owner, repo)

    def on_click(self, *_) -> None:
        """Event handler for treeview item click event."""

        item = self.tree.selection()[0]
        number = self.tree.item(item, "values")[0]
        self.open_editor(number)

    def open_editor(self, number: str) -> None:
        """Opens the issue viewer for the selected issue."""

        self.base.editorsmanager.add_editor(
            PRViewer(self.base.editorsmanager, self.prs[number])
        )

    def fetch(self) -> typing.List[dict]:
        """Fetches prs from the current repository."""

        response = requests.get(self.url)
        if response.status_code != 200:
            self.base.notifications.error(
                f"Failed to fetch PRs from {self.owner}/{self.repo}"
            )
            self.base.logger.error(
                f"Failed to fetch prs from {self.owner}/{self.repo}: {response.reason} {response.text}"
            )
            return

        prs = json.loads(response.text)
        self.prs = {str(pr["number"]): pr for pr in prs}
        if not prs:
            self.tree.insert("", tk.END, text="No open pull requests")
            return

        self.prs_actionset.update(
            [
                (
                    f"{pr['title']} #{pr['number']}",
                    lambda *_, number=pr["number"]: self.open_editor(number),
                )
                for pr in prs
            ]
        )
        prs = ((f"{pr['title']} #{pr['number']}", pr["number"]) for pr in prs)

        self.tree.delete(*self.tree.get_children())
        for pr in prs:
            self.tree.insert("", tk.END, text=pr[0], values=(pr[1],))



from __future__ import annotations

import os
import queue
import re
import threading
import tkinter as tk
import typing
from collections import deque
from tkinter.messagebox import askokcancel

import chardet
import tarts as lsp
from editorconfig import EditorConfigError
from editorconfig import get_properties as get_editorconfig

from biscuit.common.minclosestdict import MinClosestKeyDict
from biscuit.common.textindex import TextIndex
from biscuit.language.data import Diagnostic

if typing.TYPE_CHECKING:
    from biscuit.language.data import (
        WorkspaceEdits,
        HoverResponse,
        Jump,
        Completions,
        Diagnostic,
    )
    from . import TextEditor

from biscuit.common import textutils
from biscuit.common.ui import Text as BaseText

from ..comment_prefix import get_comment_prefix
from .highlighter import Highlighter

BRACKET_MAP = {"(": ")", "{": "}", "[": "]"}
BRACKET_MAP_REV = {v: k for k, v in BRACKET_MAP.items()}
OPENING_BRACKETS = ("(", "{", "[")
CLOSING_BRACKETS = (")", "}", "]")


class Text(BaseText):
    """Improved Text widget"""

    def __init__(
        self,
        master: TextEditor,
        path: str = None,
        exists: bool = True,
        minimalist: bool = False,
        standalone: bool = False,
        language: str = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: TextEditor = master
        self.path = path
        self.filename = os.path.basename(path) if path else None
        self.encoding = "utf-8"
        self.eol = "CRLF"
        self.exists = exists
        self.minimalist = minimalist
        self.standalone = standalone
        self.language = language
        self.language_alias = ""
        self.unsupported = False
        self.relative_line_numbers = self.base.relative_line_numbers

        self.ctrl_down = False
        self.buffer_size = 4096
        self.bom = True
        self.current_word = None
        self.words: list[str] = []
        self.lsp: bool = False
        self.current_indent_level = 0
        self.insert_final_newline = False

        self.hover_after = None
        self.last_hovered = None
        self.tab_spaces = self.base.tab_spaces

        try:
            self.editorconfig = get_editorconfig(self.path)

            self.insert_final_newline = self.editorconfig.get(
                "insert_final_newline", False
            )
            self.eol = self.editorconfig.get("end_of_line", "CRLF")
            self.encoding = self.editorconfig.get("charset", "utf-8")
            self.tab_spaces = int(self.editorconfig.get("indent_size", 4))
        except EditorConfigError:
            self.editorconfig = {}

        # self.last_change = Change(None, None, None, None, None)
        self.highlighter = Highlighter(self, language)
        if not self.standalone and not self.minimalist:
            self.base.statusbar.on_open_file(self)
            self.autocomplete = self.base.autocomplete
            self.definitions = self.base.peek
            self.hover = self.base.hover

        self.diagnostics = MinClosestKeyDict()

        self.focus_set()
        self.config_tags()
        self.create_proxy()
        self.config_bindings()
        self.update_idletasks()
        self.comment_prefix = get_comment_prefix(self.language.lower())
        tab_width = self.base.settings.font.measure(" " * self.tab_spaces)
        self.configure(
            tabs=(tab_width,),
            blockcursor=self.base.block_cursor,
            wrap=tk.NONE,
            relief=tk.FLAT,
            highlightthickness=0,
            bd=0,
            **self.base.theme.editors.text,
        )

        # modified event
        self.clear_modified_flag()
        self._user_edit = True
        self._edit_stack = []
        self._edit_stack_index = -1

    def config_tags(self):
        self.indentguide_stipple = self.base.resources.indent_guide

        self.tag_config(tk.SEL, background=self.base.theme.editors.selection)
        self.tag_config(
            "hyperlink",
            foreground=self.base.theme.editors.hyperlink,
            underline=True,
        )

        self.tag_config("hint", underline=True, underlinefg="gray")
        self.tag_config("information", underline=True)
        self.tag_config("warning", underline=True, underlinefg="yellow")
        self.tag_config("error", underline=True, underlinefg="red")

        self.tag_config("found", background=self.base.theme.editors.found)
        self.tag_config("foundcurrent", background=self.base.theme.editors.foundcurrent)
        self.tag_config("currentword", background=self.base.theme.editors.currentword)
        self.tag_config(
            "currentline",
            background=self.base.theme.editors.currentline,
        )
        self.tag_config("hover", background=self.base.theme.editors.hovertag)

        self.tag_configure(
            "indent_guide",
            bgstipple=f"@{self.indentguide_stipple}",
            background=self.base.theme.border,
        )
        self.tag_configure(
            "current_indent_guide",
            bgstipple=f"@{self.indentguide_stipple}",
            background=self.base.theme.secondary_foreground,
        )

        self.tag_raise(tk.SEL, "hover")
        self.tag_raise(tk.SEL, "currentline")
        self.tag_raise(tk.SEL, "currentword")
        self.tag_raise("currentword", "currentline")

        self.tag_config(
            "activebracket", background=self.base.theme.editors.activebracket
        )
        self.tag_config("red", foreground="red")
        for i in self.base.theme.editors.bracket_colors:
            self.tag_config(i, foreground=f"#{i}")

    def config_bindings(self):
        self.bind("<KeyRelease>", self.key_release_events)

        self.bind("<Control-f>", self.open_find_replace)
        self.bind("<Control-g>", lambda _: self.base.palette.show(":"))
        self.bind("<Control-Left>", lambda _: self.handle_ctrl_hmovement())
        self.bind("<Control-Right>", lambda _: self.handle_ctrl_hmovement(True))
        self.bind("<Control-Shift-Left>", lambda _: self.handle_ctrl_shift_hmovement())
        self.bind(
            "<Control-Shift-Right>", lambda _: self.handle_ctrl_shift_hmovement(True)
        )

        self.bind("<Shift-Alt-Up>", self.event_copy_line_up)
        self.bind("<Shift-Alt-Down>", self.event_copy_line_down)
        self.bind("<Alt-Up>", self.event_move_line_up)
        self.bind("<Alt-Down>", self.event_move_line_down)

        self.bind("<Return>", self.enter_key_events)
        self.bind("<Tab>", self.tab_key_events)
        self.bind("<Shift-Tab>", self.dedent_selection)

        # undo-redo
        self.bind_all("<<Modified>>", self._been_modified)

        # pair completion
        self.bind("<parenleft>", self.open_bracket)
        self.bind("<braceleft>", self.open_bracket)
        self.bind("<bracketleft>", self.open_bracket)

        self.bind("<parenright>", self.close_bracket)
        self.bind("<braceright>", self.close_bracket)
        self.bind("<bracketright>", self.close_bracket)

        self.bind("<apostrophe>", self.complete_pair)
        self.bind("<quotedbl>", self.complete_pair)
        self.bind("<BackSpace>", self.remove_pair)

        if self.minimalist or self.standalone:
            return

        self.bind("<Control-slash>", self.toggle_comment)

        # autocomplete
        self.bind("<FocusOut>", self.event_focus_out)
        self.bind("<Button-1>", self.event_mouse_down)
        self.bind("<Button-3>", self.base.text_editor_context_menu.show)
        self.bind("<Leave>", self.event_leave)
        self.bind("<Up>", self.autocomplete.move_up)
        self.bind("<Down>", self.autocomplete.move_down)

        # lspc
        self.bind("<Map>", self.event_mapped)
        self.bind("<F2>", self.request_rename)
        self.bind("<Unmap>", self.event_unmapped)
        self.bind("<Destroy>", self.event_destroy)
        self.bind("<Motion>", self.request_hover)
        self.bind("<<Selection>>", self.on_selection)
        self.bind("<Control-KeyPress>", lambda _: self.set_ctrl_key(True))
        self.bind("<Control-KeyRelease>", lambda _: self.set_ctrl_key(False))
        self.bind("<Control-Button-1>", self.request_references)
        self.bind("<Control-comma>", self.request_definition)
        self.bind(
            "<Control-period>",
            lambda _: self.base.language_server_manager.request_completions(self),
        )

        self.tag_bind("error", "<Enter>", lambda _: self.diagnostic_hover(1))
        self.tag_bind("warning", "<Enter>", lambda _: self.diagnostic_hover(2))
        self.tag_bind("information", "<Enter>", lambda _: self.diagnostic_hover(3))
        self.tag_bind("hint", "<Enter>", lambda _: self.diagnostic_hover(4))

        self.tag_bind("error", "<Leave>", self.base.diagnostic.hide)
        self.tag_bind("warning", "<Leave>", self.base.diagnostic.hide)
        self.tag_bind("information", "<Leave>", self.base.diagnostic.hide)
        self.tag_bind("hint", "<Leave>", self.base.diagnostic.hide)

    def key_release_events(self, event: tk.Event):
        self._user_edit = True

        match event.keysym.lower():
            case (
                "button-2"
                | "backspace"
                | "escape"
                | "control_l"
                | "control_r"
                | "space"
                | "return"
                | "tab"
            ):
                self.hide_autocomplete()
            case "right" | "left":
                if self.autocomplete.active:
                    self.update_completions()

            case (
                "up"
                | "down"
                | "shift_l"
                | "shift_r"
                | "alt_l"
                | "alt_r"
                | "meta_l"
                | "meta_r"
                | "shift"
                | "alt"
                | "meta"
            ):
                pass
            case "braceleft" | "bracketleft" | "parenleft" | "apostrophe" | "quotedbl":
                pass

            # auto space after : and ,
            case ":" | ",":
                self.insert(tk.INSERT, " ")
            case _:
                if self.lsp:
                    self.request_autocomplete(event)
                else:
                    self.show_autocomplete(event)

        self.update_words_list()

    def diagnostic_hover(self, severity: int) -> str:
        if pos := self.get_mouse_pos():
            message, start = self.diagnostics[pos]
            self.base.diagnostic.show(self, start, message, severity)

    def update_indent_guides(self) -> None:
        if self.minimalist:
            return

        self.tag_remove("indent_guide", "1.0", "end")
        self.tag_remove("current_indent_guide", "1.0", "end")
        lines = self.get("1.0", "end-1c").split("\n")

        self.current_indent_level = self.get_current_indent_level() - 1

        for line_number, line in enumerate(lines, start=1):
            indent_level = self.calculate_indent_level(line)
            if indent_level > 0:
                self.add_indent_guide(line_number, indent_level)

    def calculate_indent_level(self, line: str) -> int:
        indent = len(line) - len(line.lstrip())
        return indent // self.tab_spaces

    def add_indent_guide(self, line_number: int, indent_level: int) -> None:
        for level in range(indent_level):
            start_index = f"{line_number}.{level * self.tab_spaces - 1}"
            end_index = f"{line_number}.{level * self.tab_spaces + 1}"
            self.tag_add(
                (
                    "current_indent_guide"
                    if level == self.current_indent_level
                    else "indent_guide"
                ),
                start_index,
                end_index,
            )

    def get_current_indent_level(self) -> int:
        prev = self.get("insert-1l linestart", "insert-1l lineend")
        line = self.get("insert linestart", "insert lineend")
        next_line = self.get("insert+1l linestart", "insert+1l lineend")

        return max(
            self.calculate_indent_level(prev),
            self.calculate_indent_level(line),
            self.calculate_indent_level(next_line),
        )

    # TODO this wont work properly
    # write a custom ast, parser for bracket matching
    def highlight_current_brackets(self):
        self.tag_remove("activebracket", "1.0", tk.END)
        i = self.index(tk.INSERT)
        start, end = None, None

        try:
            if start := self.search(
                r"[\{\[\(\)\]\}]", i, i + " linestart", backwards=True, regexp=True
            ):
                if counter := BRACKET_MAP.get(self.get(start, start + "+1c")):
                    end = self.search(counter, i, stopindex=i + " lineend")
        except tk.TclError:
            pass

        if not end:
            return

        self.tag_add("activebracket", start, start + "+1c")
        self.tag_add("activebracket", end, end + "+1c")

    def refresh_wrap(self):
        self.config(wrap=tk.WORD if self.base.wrap_words else tk.NONE)

    def open_bracket(self, e: tk.Event):
        text = self.get("1.0", "insert")
        i = 0
        for ch in text:
            if ch in OPENING_BRACKETS:
                i += 1
            elif ch in CLOSING_BRACKETS:
                if i > 0:
                    i -= 1
        self.complete_pair(e, self.base.theme.editors.bracket_colors[(i % 3)])

        return "break"

    def close_bracket(self, e: tk.Event):
        text = self.get("1.0", "insert")
        i = -1
        stack = []
        for ch in text:
            if ch in OPENING_BRACKETS:
                i += 1
                stack.append(BRACKET_MAP[ch])
            elif ch in CLOSING_BRACKETS:
                if i > -1:
                    i -= 1
                    if stack[-1] == ch:
                        stack.pop()
        if stack and stack[-1] == e.char:
            # skip if next character is what we are closing
            if c := self.get("insert", "insert+1c"):
                if c == e.char:
                    self.mark_set(tk.INSERT, "insert+1c")
                    return "break"

        # TODO; coloring not done right. whatever color the last bracket has that color is spread to the next characters
        #     self.insert(tk.INSERT, stack.pop(), self.base.theme.editors.bracket_colors[(i%3)] if i > -1 else 'red')
        # else:
        self.insert(tk.INSERT, e.char, "red")
        return "break"

    def complete_pair(self, e: tk.Event, tag=None):
        end = {"(": ")", "{": "}", "[": "]", '"': '"', "'": "'"}.get(e.char)
        char = e.char
        # if there is selection, surround the selection with character
        if self.tag_ranges(tk.SEL):
            if tag:
                self.insert(tk.SEL_LAST, end, tag)
                self.insert(tk.SEL_FIRST, char, tag)
            else:
                self.insert(tk.SEL_LAST, end)
                self.insert(tk.SEL_FIRST, char)
            return "break"

        if char in ('"', "'") and self.get("insert-1c", "insert").strip():
            return

        # if there is no selection, insert the character and move cursor inside the pair
        # TODO; coloring is not done properly
        # if tag:
        #     self.insert(tk.INSERT, char + end, tag)
        # else:
        self.insert(tk.INSERT, char + end)
        self.mark_set(tk.INSERT, "insert-1c")
        return "break"

    def remove_pair(self, _: tk.Event):
        if self.tag_ranges(tk.SEL):
            self.delete(tk.SEL_FIRST, tk.SEL_LAST)
            return "break"

        if not self.get("insert-1c", "insert+1c") in ["()", "[]", "{}", '""', "''"]:
            return

        self.delete("insert-1c", "insert+1c")
        return "break"

    def hide_autocomplete(self, *_):
        if self.minimalist:
            return

        self.autocomplete.hide()

    def show_autocomplete(self, event: tk.Event):
        if self.minimalist or not self.current_word or event.keysym in ["Down", "Up"]:
            return
        if not self.current_word.strip().isalpha() or self.current_word.strip() != ".":
            self.hide_autocomplete()

        self.autocomplete.show(self)
        self.update_completions()

    def update_words_list(self, *_):
        if self.minimalist or self.lsp:
            return

        try:
            content = (
                self.get(1.0, "insert-1c wordstart-1c")
                + " "
                + self.get("insert+1c", tk.END)
            )
            self.words = list(set(re.findall(r"\w+", content)))
        except:
            pass

    def update_completions(self):
        """Helper function for `AutoComplete` popup.

        If LSP is enabled, then request completions from the language server.
        Otherwise, update the completions list with words in editor."""

        if self.minimalist:
            return

        if self.lsp:
            self.request_autocomplete(self)
        else:
            self.autocomplete.update_completions(self)

    def replace_current_word(self, new_word: str):
        """Helper function for `AutoComplete` popup.
        Replaces the current word with the chosen word."""

        if self.current_word.startswith("\n"):
            self.delete("insert-1c wordstart+1c", "insert")
        else:
            self.delete("insert-1c wordstart", "insert")
        self.insert("insert", new_word)

    def cursor_screen_location(self):
        """Helper function for `AutoComplete` popup positioning.
        Returns the screen location of the cursor."""

        pos_x, pos_y = self.winfo_rootx(), self.winfo_rooty()

        cursor = tk.INSERT
        bbox = self.bbox(cursor)
        if not bbox:
            return (0, 0)

        bbx_x, bbx_y, _, bbx_h = bbox
        return (pos_x + bbx_x - 1, pos_y + bbx_y + bbx_h)

    def cursor_wordstart_screen_location(self):
        """Helper function for `Rename` popup positioning.
        Returns the screen location of the word start at the cursor position."""

        pos_x, pos_y = self.winfo_rootx(), self.winfo_rooty()

        cursor = tk.INSERT + " wordstart"
        bbox = self.bbox(cursor)
        if not bbox:
            return (0, 0)

        bbx_x, bbx_y, _, bbx_h = bbox
        return (pos_x + bbx_x - 1, pos_y + bbx_y + bbx_h)

    def enter_key_events(self, *_):
        """Handles enter key press event.
        - If the autocomplete is active, then choose the selected item.
        - Otherwise check the indentation and insert a newline with proper indentation.
        """

        if not self.minimalist and self.autocomplete.active:
            self.autocomplete.choose(self)
            return "break"

        return self.check_indentation()

    def tab_key_events(self, *_):
        """Handles tab key press event.
        - If the autocomplete is active, then choose the selected item.
        - If there is a selection, then indent the selected text."""

        if not self.minimalist and self.autocomplete.active:
            self.autocomplete.choose(self)
            return "break"

        if self.tag_ranges(tk.SEL):
            return self.indent_selection()

    def comment_selection(self):
        """Helper function of toggle_comment().
        Comments selected lines with the comment prefix of the language"""

        sel_first = self.index(tk.SEL_FIRST)
        sel_last = self.index(tk.SEL_LAST)
        start_line = int(float(sel_first))
        end_line = int(float(sel_last))

        for line in range(start_line, end_line + 1):
            # skip empty lines, they won't be commented
            if not self.get(f"{line}.0", f"{line}.0 lineend").strip():
                continue

            self.insert(f"{line}.0", f"{self.comment_prefix} ")

        self.tag_remove(tk.SEL, "1.0", tk.END)
        self.tag_add(tk.SEL, sel_first, sel_last)
        return "break"

    def uncomment_selection(self):
        """Helper function of toggle_comment().
        Uncomments selected lines with the comment prefix of the language"""

        sel_first = self.index(tk.SEL_FIRST)
        sel_last = self.index(tk.SEL_LAST)
        start_line = int(float(sel_first))
        end_line = int(float(sel_last))

        for line in range(start_line, end_line + 1):
            # delete comment prefix with the trailing space
            if (
                self.get(f"{line}.0", f"{line}.{len(self.comment_prefix)+1}")
                == f"{self.comment_prefix} "
            ):
                self.delete(f"{line}.0", f"{line}.{len(self.comment_prefix)+1}")
            # trailing space not detected, delete the comment prefix
            elif (
                self.get(f"{line}.0", f"{line}.{len(self.comment_prefix)}")
                == f"{self.comment_prefix}"
            ):
                self.delete(f"{line}.0", f"{line}.{len(self.comment_prefix)}")

        self.tag_remove(tk.SEL, "1.0", tk.END)
        self.tag_add(tk.SEL, sel_first, sel_last)
        return "break"

    def toggle_comment(self, *_):
        """Toggles comments on selected lines with the comment prefix of the language"""

        if not (self.comment_prefix and self.tag_ranges(tk.SEL)):
            return "break"

        sel_first = f"{self.index(tk.SEL_FIRST)} linestart"
        sel_last = f"{self.index(tk.SEL_LAST)} lineend"

        # if all the selected *non-empty* lines start with comment prefix, then proceed to uncomment
        if all(
            (
                i.startswith(self.comment_prefix)
                for i in self.get(sel_first, sel_last).split("\n")
                if i.strip()
            )
        ):
            return self.uncomment_selection()

        # otherwise comment the selected lines
        return self.comment_selection()

    def dedent_selection(self, _):
        """Dedent the selected text by removing tabs or spaces at the start of each line"""

        sel_first = self.index(tk.SEL_FIRST)
        sel_last = self.index(tk.SEL_LAST)
        start_line = int(float(sel_first))
        end_line = int(float(sel_last))

        for line in range(start_line, end_line + 1):
            if (
                self.get(f"{line}.0", f"{line}.1") == "\t"
                or self.get(f"{line}.0", f"{line}.{self.tab_spaces}")
                == " " * self.tab_spaces
            ):
                self.delete(f"{line}.0", f"{line}.1")

        self.tag_remove(tk.SEL, "1.0", tk.END)
        self.tag_add(tk.SEL, sel_first, sel_last)
        return "break"

    def indent_selection(self):
        """Indent the selected text by adding tabs at the start of each line"""

        sel_first = self.index(tk.SEL_FIRST)
        sel_last = self.index(tk.SEL_LAST)
        start_line = int(float(sel_first))
        end_line = int(float(sel_last))

        for line in range(start_line, end_line + 1):
            self.insert(f"{line}.0", "\t")

        self.tag_remove(tk.SEL, "1.0", tk.END)
        self.tag_add(tk.SEL, sel_first, sel_last)
        return "break"

    def refresh(self):
        self.highlighter.highlight()
        self.highlight_current_word()

        if self.minimalist or self.standalone:
            return

        self.current_word = self.get("insert-1c wordstart", "insert")
        self.base.language_server_manager.request_outline(self)
        self.highlight_current_line()
        self.highlight_current_brackets()
        self.update_indent_guides()

        # TODO send only portions of text on change to the LSPServer
        # current solution is not scalable, and will cause lag on large files

        # commenting this out for now, as the way of handling changes is not properly done
        # the changes shall be recorded before they are actually made, but we can't do that from here
        # this requires writing a custom change handler for text widget
        # eg. in the 'delete' command below, i can get indices like ('sel.start', 'sel.end') but not actual indices
        # these are useless as if we try to index() these indices we will get indices from the modified content,
        # not the old one. and since the modified content no longer have sel, this will throw an error

        # if args[0] == tk.INSERT:
        #     start_index = self.get_cursor_pos()

        #     end_index = self.index(tk.INSERT + f"+{len(args[2])}c")
        #     self.last_change.update(
        #         start=[int(i) for i in start_index.split('.')],
        #         old_end=[int(i) for i in start_index.split('.')],
        #         new_end=[int(i) for i in end_index.split('.')],
        #         old_text='',
        #         new_text=args[2]
        #     )

        # elif args[0] == 'delete':
        #     start_index = self.get_cursor_pos()
        #     if len(args) == 2:
        #         # is one char deleted
        #         # ('delete', 'insert-1c')
        #         start_index = self.index(tk.INSERT + f"-1c")
        #         end_index = self.get_cursor_pos()
        #         print(start_index, self.get(start_index, end_index), "------------------------------")
        #     else:
        #         # there must be selection then
        #         # ('delete', 'sel.first', 'sel.last')
        #         start_index = self.index(tk.SEL_FIRST)
        #         end_index = self.index(tk.SEL_LAST)
        #         print(start_index, self.get(start_index, end_index), "------------------------------")

        #     self.last_change.update(
        #         start=[int(i) for i in start_index.split('.')],
        #         old_end=[int(i) for i in start_index.split('.')],
        #         new_end=[int(i) for i in end_index.split('.')],
        #         old_text='',
        #         new_text=''
        #     )
        # elif args[0] == 'replace':
        #     start_index = self.get_cursor_pos()
        #     end_index = self.index(tk.INSERT + f"+{len(args[3])}c")
        #     self.last_change.update(
        #         start=[int(i) for i in start_index.split('.')],
        #         old_end=[int(i) for i in start_index.split('.')],
        #         new_end=[int(i) for i in end_index.split('.')],
        #         old_text=args[3],
        #         new_text=args[2]
        #     )
        # else:
        #     return
        # self.base.languageservermanager.content_changed(self)

    def is_identifier(self, text: str) -> str:
        return bool(re.match("^[a-zA-Z_][a-zA-Z0-9_]*$", text))

    def set_ctrl_key(self, flag):
        self.ctrl_down = flag

    def clear_goto_marks(self):
        self.tag_remove("hyperlink", 1.0, tk.END)

    def request_definition(self, from_menu=False, *_):
        if not from_menu and (not self.lsp or not self.last_hovered):
            return

        self.base.language_server_manager.request_goto_definition(self)

    def request_references(self, from_menu=False, *_):
        if not from_menu and (not self.lsp or not self.last_hovered):
            return

        self.base.language_server_manager.request_references(self)
        return "break"

    def request_hover(self, _):
        if not self.lsp or self.tag_ranges(tk.SEL):
            return

        index = self.index(tk.CURRENT)
        start, end = index + " wordstart", index + " wordend"
        word = self.get(start, end)
        if word.startswith("\n"):
            start = index + " linestart"
        word = word.strip()

        self.clear_goto_marks()

        if any(token.startswith("Token.Keyword") for token in self.tag_names(start)):
            return

        if word and self.is_identifier(word):
            if self.ctrl_down:
                self.tag_add("hyperlink", start, end)
        else:
            # TODO hide with a delay, cancel hiding if mouse is on top of hover popup
            self.hover.hide()
            self.tag_remove("hover", 1.0, tk.END)
            self.last_hovered = None
            return

        if self.last_hovered == word:
            return

        self.last_hovered = word
        self.tag_remove("hover", 1.0, tk.END)
        self.tag_add("hover", start, end)

        # TODO delayed hovers
        # if self.hover_after:
        #     self.after_cancel(self.hover_after)

        # self.after(500, ...)

        self.base.language_server_manager.request_hover(self)

    def request_autocomplete(self, _):
        if self.minimalist or not self.lsp:
            return

        if self.is_identifier(self.current_word) or self.current_word.strip() == ".":
            return self.base.language_server_manager.request_completions(self)

        self.hide_autocomplete()

    def request_rename(self, *_):
        self.base.rename.show(self)

    def lsp_show_autocomplete(self, response: Completions) -> None:
        self.autocomplete.lsp_update_completions(self, response.completions)

    def lsp_diagnostics(self, response: list[Diagnostic]) -> None:
        self.tag_remove("error", 1.0, tk.END)
        self.tag_remove("warning", 1.0, tk.END)
        self.tag_remove("information", 1.0, tk.END)
        self.tag_remove("hint", 1.0, tk.END)

        self.diagnostics.clear()

        for i in response:
            self.diagnostics[i.start] = i.message

            match i.severity:
                case 1:
                    self.tag_add("error", i.start, i.end)
                case 2:
                    self.tag_add("warning", i.start, i.end)
                case 4:
                    self.tag_add("hint", i.start, i.end)
                case _:
                    self.tag_add("information", i.start, i.end)

    def lsp_goto_definition(self, response: Jump) -> None:
        if not response.locations:
            return

        if len(response.locations) == 1:
            return self.base.goto_location(
                response.locations[0].file_path, response.locations[0].start
            )

        self.definitions.show(self, response)

    def lsp_hover(self, response: HoverResponse) -> None:
        if not (response.text or response.docs):
            return self.hover.hide()

        try:
            self.hover.show(self, response)
        except Exception as e:
            print(e)

    def lsp_rename(self, changes: WorkspaceEdits):
        for i in changes.edits:
            try:
                self.base.open_workspace_edit(i.file_path, i.edits)
            except:
                # this indeed is a darn task to do so not surprised
                pass

    def set_tab_size(self, size: int) -> None:
        if not size:
            return

        tab_width = self.base.settings.font.measure(" " * size)
        self.configure(tabs=(tab_width,))

    def set_block_cursor(self, flag: bool) -> None:
        self.configure(blockcursor=flag)

    def toggle_relative_numbering(self) -> None:
        self.relative_line_numbers = not self.relative_line_numbers
        self.master.on_change()

    def get_cursor_pos(self):
        return self.index(tk.INSERT)

    def get_mouse_pos(self):
        return self.index(tk.CURRENT)

    def get_current_word(self) -> str:
        """Returns current word cut till the cursor"""

        return self.current_word.strip()

    def get_current_fullword(self) -> str | None:
        """Returns current word uncut and fully"""

        index = self.index(tk.INSERT)
        start, end = index + " wordstart", index + " wordend"
        word = self.get(start, end).strip()
        return word if self.is_identifier(word) else None

    def handle_ctrl_hmovement(self, delta=False):
        if delta:
            self.move_to_next_word()
        else:
            self.move_to_previous_word()

        return "break"

    def handle_ctrl_shift_hmovement(self, delta=False):
        if delta:
            self.select_to_next_word()
        else:
            self.select_to_previous_word()

        return "break"

    def select_to_next_word(self):
        self.tag_add(tk.SEL, "insert wordstart", "insert+1c wordend")
        self.mark_set(tk.INSERT, "insert+1c wordend")

    def select_to_previous_word(self):
        self.tag_add(tk.SEL, "insert-1c wordstart", "insert wordend")
        self.mark_set(tk.INSERT, "insert-1c wordstart")

    def move_to_next_word(self):
        self.mark_set(tk.INSERT, self.index("insert+1c wordend"))

    def move_to_previous_word(self):
        self.mark_set(tk.INSERT, self.index("insert-1c wordstart"))

    def on_selection(self, *args):
        self.tag_remove("hover", 1.0, tk.END)

    def update_current_indent(self):
        line = self.get("insert linestart", "insert lineend")
        match = re.match(r"^(\s+)", line)
        self.current_indent = len(match.group(0)) if match else 0

    def update_current_line(self):
        self.current_line = self.get("insert linestart", "insert lineend")
        return self.current_line

    def add_newline(self, count=1):
        self.insert(tk.INSERT, "\n" * count)

    def check_indentation(self, *args):
        self.update_current_indent()
        if self.update_current_line():
            if self.current_line[-1] in ["{", "[", ":", "("]:
                self.current_indent += self.tab_spaces
            elif self.current_line[-1] in ["}", "]", ")"]:
                self.current_indent -= self.tab_spaces

            self.add_newline()
            self.insert(tk.INSERT, " " * self.current_indent)

            self.update_current_indent()

            return "break"

    def multi_selection(self, *args):
        # TODO: multi cursor editing
        return "break"

    def open_find_replace(self, *_):
        self.base.findreplace.show(self)

    def detect_encoding(self, file_path):
        if not self.exists:
            return "utf-8"

        try:
            with open(file_path, "rb") as file:
                bomstring = file.read(8)
                detected = chardet.detect(bomstring)
                encoding = detected["encoding"].lower()
                if encoding == "ascii":
                    return "utf-8"

                return encoding
        except Exception as e:
            # empty file
            return "utf-8"

    def change_eol(self, eol: str):
        if not self.exists:
            text = self.get_all_text()
            if not text.strip():
                self.eol = eol
                self.base.statusbar.on_open_file(self)
                return
        self.encoding = self.encoding or self.detect_encoding(self.path)
        if self.path and self.exists:
            self.clear()
            file = open(
                self.path, "r", encoding=self.encoding, buffering=self.buffer_size
            )
            self.queue = queue.Queue()
            threading.Thread(target=self.read_file, args=(file,)).start()
            self.process_queue(eol=eol)
        else:
            self.load_text(text, eol=eol)
        self.eol = eol
        self.base.statusbar.on_open_file(self)

    def reopen(self, encoding=None, eol=None) -> None:
        if not self.path or not self.exists:
            return

        self.clear()
        self.highlighter.clear()
        try:
            self.encoding = encoding or self.encoding
            self.eol = eol or self.eol

            file = open(
                self.path,
                "r",
                buffering=self.buffer_size,
                encoding=self.encoding,
                newline=self.eol,
            )

            self.queue = queue.Queue()
            threading.Thread(target=self.read_file, args=(file,)).start()
            self.process_queue()
        except Exception as e:
            print(e)
            if self.exists:
                self.master.unsupported_file()

        self.refresh()
        self.base.statusbar.on_open_file(self)

    def load_file(self):
        if not self.path or not self.exists:
            return

        if not textutils.is_text_file(self.path):
            if not askokcancel(
                "Binary File",
                "This file is a binary file, do you want to open it anyway?",
            ):
                # show unsupported file message and terminate content loading process
                return self.master.unsupported_file()
            else:
                # just set the flag to True so that we don't cache editor content
                self.master.unsupported = True
                self.unsupported = True

        self.clear()
        try:
            self.encoding = self.detect_encoding(self.path)
            file = open(
                self.path, "r", encoding=self.encoding, buffering=self.buffer_size
            )
            self.eol = textutils.get_default_newline()

            self.queue = queue.Queue()
            threading.Thread(target=self.read_file, args=(file,), daemon=True).start()
            self.process_queue()
        except Exception as e:
            print(e)
            if self.exists:
                self.master.unsupported_file()

        if not self.standalone:
            self.base.statusbar.on_open_file(self)

    def load_new_file(self, path: str):
        self.path = path
        self.load_file()
        self.highlighter.detect_language()
        self.highlighter.highlight()

    def load_text(self, text: str = "", eol: str = ""):
        self.clear()

        def write_with_buffer():
            buffer = deque(maxlen=self.buffer_size)
            for char in text:
                buffer.append(char)
                if len(buffer) >= self.buffer_size:
                    chunk = "".join(buffer)
                    if eol:
                        chunk.replace(self.eol or textutils.get_default_newline(), eol)
                    self.write(chunk)
                    self.update()
                    buffer.clear()
            if buffer:
                chunk = "".join(buffer)
                self.write(chunk)
                self.update()
            if eol:
                self.eol = eol

            if not self.standalone:
                self.base.statusbar.on_open_file(self)

        threading.Thread(target=write_with_buffer, daemon=True).start()

    def read_file(self, file: typing.TextIO):
        while True:
            try:
                chunk = file.read(self.buffer_size)
            except UnicodeDecodeError as e:
                print(e)
                self.master.unsupported_file()
                return
            if not chunk:
                file.close()
                self.queue.put(None)  # Signal the end of reading
                break
            self.queue.put(chunk)

    def process_queue(self, eol: str = None):
        try:
            while True:
                chunk = self.queue.get_nowait()
                if chunk is None:
                    # Finished loading file -- reached EOF 🚧
                    try:
                        self.master.on_change()
                        self.master.on_scroll()
                        self.update_idletasks()
                        self.master.file_loaded()
                        self.focus_set()
                    except Exception:
                        pass
                    break
                if eol:
                    chunk.replace(self.eol or textutils.get_default_newline(), eol)

                try:
                    self.write(chunk)
                    self.update()
                    self.master.on_scroll()
                except Exception:
                    # editor was closed during file load
                    return
        except queue.Empty:
            # If the queue is empty, schedule the next check after a short delay
            self.master.after(100, self.process_queue)

    def custom_get(self, start: str, end: str) -> str:
        """Ignore the text that is tagged with 'ignore_tag' and return the rest of the text."""

        content = self.get(start, end)
        tag_ranges = self.tag_ranges("ignore_tag")

        for tag_start, tag_end in zip(tag_ranges[0::2], tag_ranges[1::2]):
            content = content.replace(self.get(tag_start, tag_end), "")

        return content

    def save_file(self, path=None):
        if self.insert_final_newline:
            if not self.get("end-2c", "end-1c").endswith("\n"):
                self.add_newline()

        if path:
            try:
                with open(path, "w") as fp:
                    fp.write(self.get_all_text())
            except Exception:
                return

            self.path = path
            # TODO update tab name

        try:
            with open(self.path, "w") as fp:
                fp.write(self.get_all_text())
        except Exception:
            return

    def event_focus_out(self, _: tk.Event):
        self.hide_autocomplete()
        self.hover.hide()

    def event_mouse_down(self, _: tk.Event):
        self.hide_autocomplete()
        self.hover.hide()

    def event_leave(self, _: tk.Event):
        self.hover.hide()

    def event_mapped(self, _):
        try:
            self.lsp = self.base.language_server_manager.tab_opened(self)
        except Exception as e:
            self.base.notifications.warning(
                f"{self.language} language server failed, check logs."
            )
            self.base.logger.error(f"{self.language} language server failed: {e}")
            self.lsp = False
            print(self.base.language_server_manager.langservers)

    def event_destroy(self, _):
        try:
            self.hide_autocomplete()
        except:
            # most likely because app was closed
            pass
        self.base.language_server_manager.request_removal(self)

    def event_unmapped(self, _):
        try:
            self.hide_autocomplete()
            self.base.outline.tree.update_symbols()
        except Exception:
            pass

    def copy(self, *_):
        self.event_generate("<<Copy>>")

    def cut(self, *_):
        self.event_generate("<<Cut>>")

    def paste(self, *_):
        self.event_generate("<<Paste>>")

    def clear(self) -> None:
        """Clear the entire text content"""

        self.delete(1.0, tk.END)

    def goto(self, position: str) -> None:
        """Moves cursor to the position passed as argument"""

        self.move_cursor(position)
        self.see(position)
        self.focus_set()

    def goto_line(self, line: str) -> None:
        """Moves cursor to the line passed as argument"""

        line = f"{line}.0"
        self.move_cursor(line)
        self.see(line)
        self.focus_set()

    def event_copy_line_up(self, *_) -> None:
        "copies the line cursor is in below"
        line = self.line
        next_line = str(int(line) + 1)
        self.insert(f"{next_line}.0", self.get(f"{line}.0", f"{line}.end"))
        return "break"

    def event_copy_line_down(self, *_) -> None:
        "copies the line cursor is in above"
        line = self.line
        prev_line = str(int(line) - 1)
        self.insert(f"{prev_line}.end", self.get(f"{line}.0", f"{line}.end"))
        return "break"

    def event_delete_line(self, *_) -> None:
        "deletes the line cursor is in"
        line = self.line
        self.delete(f"{line}.0", f"{line}.end")

    def event_move_line_up(self, *_) -> None:
        "moves the line cursor is in below"
        line = self.line
        next_line = str(int(line) + 1)
        self.insert(f"{next_line}.0", self.get(f"{line}.0", f"{line}.end"))
        self.delete(f"{line}.0", f"{line}.end")
        return "break"

    def event_move_line_down(self, *_) -> None:
        "moves the line cursor is in above"
        line = self.line
        prev_line = str(int(line) - 1)
        self.insert(f"{prev_line}.end", self.get(f"{line}.0", f"{line}.end"))
        self.delete(f"{line}.0", f"{line}.end")
        return "break"

    def event_duplicate_selection(self, *_) -> None:
        "duplicates the current selection"
        self.insert(tk.INSERT, self.selection)

    def write(self, text, *args):
        self.insert(tk.END, text, *args)

    def insert_newline(self, *args):
        self.write("\n", *args)

    def get_begin(self):
        return "1.0"

    def get_end(self):
        return self.index(tk.END + "-1c")

    def get_all_text(self):
        return self.get(1.0, tk.END)

    @property
    def selection(self) -> str:
        try:
            return self.selection_get()
        except Exception:
            return ""

    @property
    def line(self):
        try:
            return int(self.index(tk.INSERT).split(".")[0])
        except:
            return 1

    @property
    def column(self):
        try:
            return int(self.index(tk.INSERT).split(".")[1]) + 1
        except:
            return 1

    @property
    def position(self):
        try:
            lc = self.index(tk.INSERT).split(".")
            return [lc[0], int(lc[1]) + 1]
        except:
            return [1, 1]

    def scroll_to_end(self):
        self.mark_set(tk.INSERT, tk.END)
        self.see(tk.INSERT)

    def scroll_to_start(self):
        self.mark_set(tk.INSERT, 1.0)
        self.see(tk.INSERT)

    def scroll_to_line(self, line):
        self.mark_set(tk.INSERT, line)
        self.see(tk.INSERT)

    def set_wrap(self, flag=True):
        if flag:
            self.configure(wrap=tk.WORD)
        else:
            self.configure(wrap=tk.NONE)

    def set_active(self, flag=True):
        if flag:
            self.configure(state=tk.NORMAL)
        else:
            self.configure(state=tk.DISABLED)

    def show_unsupported_dialog(self):
        self.minimalist = True
        self.unsupported = True
        self.highlighter.lexer = None
        self.set_wrap(True)
        self.configure(font=("Arial", 10), padx=10, pady=10)
        self.write(
            "This file is not displayed in this editor because it is either binary or uses an unsupported text encoding."
        )
        self.set_active(False)

    def move_cursor(self, position):
        self.mark_set(tk.INSERT, position)

    def clear_all_selection(self):
        self.tag_remove(tk.SEL, 1.0, tk.END)

    def highlight_current_line(self, *_):
        self.tag_remove("currentline", 1.0, tk.END)
        if self.minimalist or self.tag_ranges(tk.SEL):
            return

        line = int(self.index(tk.INSERT).split(".")[0])
        start = str(float(line))
        end = str(float(line + 1))
        self.tag_add("currentline", start, end)

    def select_line(self, line):
        self.clear_all_selection()

        line = int(line.split(".")[0])
        start = str(float(line))
        end = str(float(line + 1))
        self.tag_add(tk.SEL, start, end)

        self.move_cursor(end)

    def highlight_current_word(self):
        self.tag_remove("currentword", 1.0, tk.END)
        if self.tag_ranges(tk.SEL):
            return

        if word := re.findall(r"\w+", self.get("insert wordstart", "insert wordend")):
            # TODO: do not highlight keywords, parts of strings, etc.
            self.highlight_pattern(f"\\y{word[0]}\\y", "currentword", regexp=True)
            # self.highlight_pattern(f"\\y{word[0]}\\y", "currentword", start="insert wordend", regexp=True)

    def highlight_pattern(self, pattern, tag, start="1.0", end=tk.END, regexp=False):
        start = self.index(start)
        end = self.index(end)

        self.mark_set("matchStart", start)
        self.mark_set("matchEnd", start)
        self.mark_set("searchLimit", end)

        self.tag_remove(tag, start, end)

        count = tk.IntVar()
        while True:
            index = self.search(
                pattern, "matchEnd", "searchLimit", count=count, regexp=regexp
            )
            if index == "" or count.get() == 0:
                break

            self.mark_set("matchStart", index)
            self.mark_set("matchEnd", f"{index}+{count.get()}c")

            self.tag_add(tag, "matchStart", "matchEnd")

    def stack_undo(self):
        if self._edit_stack_index > 0:
            self._edit_stack_index = self._edit_stack_index - 1
            self._user_edit = False
            self.clear()
            self.write(self._edit_stack[self._edit_stack_index][0][:-1])
            self.mark_set("insert", self._edit_stack[self._edit_stack_index][1])

    def stack_redo(self):
        if self._edit_stack_index + 1 < len(self._edit_stack):
            self._edit_stack_index = self._edit_stack_index + 1
            self._user_edit = False
            self.clear()
            self.write(self._edit_stack[self._edit_stack_index][0][:-1])
            self.mark_set("insert", self._edit_stack[self._edit_stack_index][1])

    def _been_modified(self, event=None):
        try:
            if self._user_edit:
                text = self.get_all_text()
                if (not self._edit_stack) or (
                    text != self._edit_stack[self._edit_stack_index][0]
                ):
                    # real modified
                    cursor_index = self.index(tk.INSERT)
                    if (self._edit_stack_index + 1) != len(self._edit_stack):
                        self._edit_stack = self._edit_stack[
                            : self._edit_stack_index + 1
                        ]
                    self._edit_stack.append([text, cursor_index])
                    self._edit_stack_index = self._edit_stack_index + 1
                    if self._edit_stack_index > 200:
                        self._edit_stack = self._edit_stack[
                            self._edit_stack_index - 50 : self._edit_stack_index + 1
                        ]
                        self._edit_stack_index = len(self._edit_stack) - 1
            if self._resetting_modified_flag:
                return
            self.clear_modified_flag()
        except:
            self.base.notifications.error("Edit stack error: please restart biscuit")

    def clear_modified_flag(self):
        self._resetting_modified_flag = True
        try:
            self.tk.call(self._w, "edit", "modified", 0)
        finally:
            self._resetting_modified_flag = False

    def create_proxy(self):
        self._orig = self._w + "_orig"
        self.tk.call("rename", self._w, self._orig)
        self.tk.createcommand(self._w, self._proxy)

    def _proxy(self, *args):
        if (
            args[0] == "get"
            and (args[1] == tk.SEL_FIRST and args[2] == tk.SEL_LAST)
            and not self.tag_ranges(tk.SEL)
        ):
            return
        if (
            args[0] == "delete"
            and (args[1] == tk.SEL_FIRST and args[2] == tk.SEL_LAST)
            and not self.tag_ranges(tk.SEL)
        ):
            return

        cmd = (self._orig,) + args
        try:
            result = self.tk.call(cmd)
        except Exception:
            return

        if args[0] in ("insert", "replace", "delete"):
            self.event_generate("<<Change>>", when="tail")
            if self.lsp:
                self.base.language_server_manager.content_changed(self)

        # if "insert" in args[0:3] and "get" in args[0:3]:
        #     print(temp)

        elif args[0:3] == ("mark", "set", "insert"):
            self.event_generate("<<Change>>", when="tail")
        elif (
            args[0:2] == ("xview", "moveto")
            or args[0:2] == ("yview", "moveto")
            or args[0:2] == ("xview", "scroll")
            or args[0:2] == ("yview", "scroll")
        ):
            self.event_generate("<<Scroll>>", when="tail")

        return result



__author__ = "cid0rz"


import random
import tkinter as tk

from biscuit.common.ui import Canvas

from .gamebase import BaseGame

WIDTH = 800
HEIGHT = 400
BALL_RADIUS = 10
PADDLE_WIDTH = 10
PADDLE_HEIGHT = 80
PADDLE_SPEED = 5


class Pong(BaseGame):
    name = "Pong!"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ball_dx = self.ball_dy = 5
        self.paddle1_dy = self.paddle2_dy = 0
        self.score1 = self.score2 = 0

        self.canvas = Canvas(
            self, width=WIDTH, height=HEIGHT + 10, **self.base.theme.editors
        )
        self.canvas.pack(pady=30)

        self.paddle1 = self.canvas.create_rectangle(
            10,
            HEIGHT / 2 - PADDLE_HEIGHT / 2,
            20,
            HEIGHT / 2 + PADDLE_HEIGHT / 2,
            outline="",
            fill=self.base.theme.biscuit,
        )
        self.paddle2 = self.canvas.create_rectangle(
            WIDTH - 20,
            HEIGHT / 2 - PADDLE_HEIGHT / 2,
            WIDTH - 10,
            HEIGHT / 2 + PADDLE_HEIGHT / 2,
            outline="",
            fill=self.base.theme.biscuit,
        )
        self.ball = self.canvas.create_oval(
            WIDTH / 2 - BALL_RADIUS,
            HEIGHT / 2 - BALL_RADIUS,
            WIDTH / 2 + BALL_RADIUS,
            HEIGHT / 2 + BALL_RADIUS,
            outline="",
            fill=self.base.theme.biscuit,
        )

        self.score1_text = self.canvas.create_text(
            WIDTH / 4,
            50,
            text=self.score1,
            fill=self.base.theme.biscuit_dark,
            font=("Fixedsys", 30, "bold"),
        )
        self.canvas.create_text(
            2 * WIDTH / 4,
            50,
            text="|",
            fill=self.base.theme.biscuit_dark,
            font=("Fixedsys", 30, "bold"),
        )
        self.score2_text = self.canvas.create_text(
            3 * WIDTH / 4,
            50,
            text=self.score2,
            fill=self.base.theme.biscuit_dark,
            font=("Fixedsys", 30, "bold"),
        )

        self.canvas.create_rectangle(
            0, 0, 1000, 10, outline="", fill=self.base.theme.biscuit_dark
        )
        self.canvas.create_rectangle(
            0, HEIGHT, 1000, HEIGHT + 10, outline="", fill=self.base.theme.biscuit_dark
        )

        self.bind_all("<KeyPress>", self.move_paddle)
        self.bind_all("<KeyRelease>", self.stop_paddle)
        self.update_game()

    def update_game(self):
        self.canvas.move(self.paddle1, 0, self.paddle1_dy)
        self.canvas.move(self.paddle2, 0, self.paddle2_dy)

        self.canvas.move(self.ball, self.ball_dx, self.ball_dy)

        self.ball_pos = self.canvas.coords(self.ball)
        self.paddle1_pos = self.canvas.coords(self.paddle1)
        self.paddle2_pos = self.canvas.coords(self.paddle2)

        # Ball collision with paddles
        if (
            self.ball_pos[0] <= PADDLE_WIDTH
            and self.paddle1_pos[1] < self.ball_pos[1] < self.paddle1_pos[3]
        ):
            self.ball_dx = abs(self.ball_dx)
        elif (
            self.ball_pos[2] >= WIDTH - PADDLE_WIDTH
            and self.paddle2_pos[1] < self.ball_pos[1] < self.paddle2_pos[3]
        ):
            self.ball_dx = -abs(self.ball_dx)

        # Ball collision with walls
        if self.ball_pos[1] <= 0 or self.ball_pos[3] >= HEIGHT:
            self.ball_dy = -self.ball_dy

        # Ball out of bounds
        if self.ball_pos[0] <= 0:
            self.score2 += 1
            self.canvas.itemconfig(self.score2_text, text=self.score2)
            self.reset_ball()
        elif self.ball_pos[2] >= WIDTH:
            self.score1 += 1
            self.canvas.itemconfig(self.score1_text, text=self.score1)
            self.reset_ball()

        self.canvas.after(15, self.update_game)

    def move_paddle(self, event: tk.Event):
        # Move paddle 1
        if event.keysym == "w":
            self.paddle1_dy = -PADDLE_SPEED
        elif event.keysym == "s":
            self.paddle1_dy = PADDLE_SPEED

        # Move paddle 2
        if event.keysym == "Up":
            self.paddle2_dy = -PADDLE_SPEED
        elif event.keysym == "Down":
            self.paddle2_dy = PADDLE_SPEED

    def stop_paddle(self, event: tk.Event):
        # Stop paddle 1
        if event.keysym in ["w", "s"]:
            self.paddle1_dy = 0

        # Stop paddle 2
        if event.keysym in ["Up", "Down"]:
            self.paddle2_dy = 0

    def reset_ball(self):
        self.canvas.coords(
            self.ball,
            WIDTH / 2 - BALL_RADIUS,
            HEIGHT / 2 - BALL_RADIUS,
            WIDTH / 2 + BALL_RADIUS,
            HEIGHT / 2 + BALL_RADIUS,
        )
        self.canvas.move(
            self.ball,
            random.choice([-1, 1]) * random.randint(2, 4),
            random.choice([-1, 1]) * random.randint(2, 4),
        )



import textwrap
import tkinter as tk

from ..icons import Icons
from .native import Frame, Label


class WrappingLabel(Label):
    """a type of Label that automatically adjusts the wrap to the size"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.bind("<Configure>", lambda e: self.config(wraplength=self.winfo_width()))


class IconLabel(Frame):
    def __init__(
        self,
        master,
        text=None,
        icon: Icons = None,
        iconside=tk.LEFT,
        fg=None,
        font=None,
        padx=5,
        pady=1,
        expandicon=True,
        highlighted=False,
        iconsize=14,
        toggle=True,
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, padx=padx, pady=pady, *args, **kwargs)
        self.bg, self.fg, self.hbg, self.hfg = (
            self.base.theme.utils.iconlabelbutton.values()
            if not highlighted
            else self.base.theme.utils.button.values()
        )
        self.fg = fg or self.fg
        self.config(bg=self.bg)
        self.text = text
        self.icon = icon
        self.codicon = self.icon

        self.toggle = toggle

        if icon:
            self.icon_label = tk.Label(
                self,
                text=self.codicon,
                anchor=tk.E,
                bg=self.bg,
                fg=self.fg,
                font=("codicon", iconsize),
            )
            self.icon_label.pack(side=iconside, fill=tk.BOTH, expand=expandicon)

        if text:
            self.text_label = tk.Label(
                self,
                text=self.text,
                anchor=tk.W,
                pady=2,
                bg=self.bg,
                fg=self.fg,
                font=font or self.base.settings.uifont,
            )
            self.text_label.pack(side=iconside, fill=tk.BOTH, expand=True)

        self.visible = False

    def toggle_icon(self) -> None:
        self.icon_label.config(text=self.codicon if not self.toggle else "    ")
        self.toggle = not self.toggle

    def change_text(self, text) -> None:
        self.text_label.config(text=text)

    def change_icon(self, icon) -> None:
        self.icon_label.config(text=icon)

    def set_pack_data(self, **kwargs) -> None:
        self.pack_data = kwargs

    def get_pack_data(self):
        return self.pack_data

    def show(self) -> None:
        if not self.visible:
            self.visible = True
            self.pack(**self.get_pack_data())

    def hide(self) -> None:
        if self.visible:
            self.visible = False
            self.pack_forget()


class TruncatedLabel(Frame):
    """Truncates long text and adds filler "..." at the end
    NOTE: Doesnt work currently"""

    def __init__(self, master, text, *args, **kwargs) -> None:
        super().__init__(master)
        self.text = text
        self.pack_propagate(False)

        self.label = Label(self, text=text, *args, **kwargs)
        self.label.pack(fill="both", expand=True)

        self.bind("<Configure>", self._wrap_text)

    def _wrap_text(self, event=None):
        width = self.winfo_width() * (len(self.text) / self.label.winfo_width())

        if width > 0:
            wrapped_text = textwrap.fill(self.text, width=width)

            if len(wrapped_text) > len(self.text):
                # Truncate the text and add "..." at the end
                truncated_text = textwrap.shorten(
                    wrapped_text, width - 3, placeholder="..."
                )
                self.label.configure(text=truncated_text)
            else:
                self.label.configure(text=wrapped_text)



from .game import *
from .gamebase import BaseGame
from .gameoflife import GameOfLife
from .minesweeper import Minesweeper
from .pong import Pong
from .ttt import TicTacToe
from .whoops import Whoops



from __future__ import annotations

import os
import typing

from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

if typing.TYPE_CHECKING:
    from .directorytree import DirectoryTree


class DirectoryTreeWatcher(PatternMatchingEventHandler):
    """Watchdog event handler for the directory tree.

    The DirectoryTreeWatcher is used to watch for changes in the directory tree.
    - The DirectoryTreeWatcher is used to update the directory tree when changes are detected.
    - The DirectoryTreeWatcher is used to reload the source control tree when changes are detected.
    """

    def __init__(self, master: DirectoryTree, tree, observe_changes) -> None:
        self.master = master
        self.base = master.base
        self.tree = tree
        self.observe_changes = observe_changes

        super().__init__(ignore_patterns=self.master.changes_ignore_dir_patterns)

        self.observer = Observer()
        self.observer.start()

    def watch(self) -> None:
        self.observer.unschedule_all()
        if self.master.path and self.observe_changes:
            self.observer.schedule(self, self.master.path, recursive=True)

    def stop_watch(self) -> None:
        self.observer.stop()

    def on_created(self, event) -> None:
        for i in self.master.search_ignore_dirs:
            if i in event.src_path:
                return

        self.master.update_path(os.path.dirname(event.src_path))
        self.base.source_control.reload_tree()

    def on_deleted(self, event) -> None:
        for i in self.master.search_ignore_dirs:
            if i in event.src_path:
                return

        self.master.update_path(os.path.dirname(event.src_path))
        self.base.source_control.reload_tree()

    def on_modified(self, event) -> None:
        # self.master.update_path(os.path.dirname(event.src_path))
        # self.base.source_control.reload_tree()
        # print('modified', event.src_path)
        ...

    def on_moved(self, event):
        for i in self.master.search_ignore_dirs:
            if i in event.src_path:
                return

        try:
            self.master.update_path(os.path.dirname(event.src_path))
            self.base.source_control.reload_tree()
            print("moved", event.src_path, event.dest_path)
        except FileNotFoundError:
            pass



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from biscuit.debugger.manager import DebuggerManager
    from biscuit.editor import TextEditor


class DebuggerBase:
    """Abstract debugger base class.
    This class should be inherited by a debugger class that implements the `run` method.

    Attributes:
        manager (DebuggerManager): the debugger manager
        base (App): the base app instance
        breakpoints (dict[str, set[int]]): map of breakpoints set for each file,
            `file_path` -> set of line numbers
        variables (Variables): the variables pane in the debug view,
            `variables.tree` is the treeview widget for the variables
        callstack (CallStack): the call stack pane in the debug view,
            `callstack.tree` is the treeview widget for the call stack
    """

    def __init__(self, manager: DebuggerManager):
        super().__init__()
        self.manager = manager
        self.base = manager.base
        self.variables = self.base.sidebar.debug.variables
        self.callstack = self.base.sidebar.debug.callstack
        self.breakpoints: dict[str, set[int]] = {}  # file_path -> set of line numbers

    def format_path(self, path: str) -> str:
        return path.format(workspace=self.base.active_directory)

    def launch_standalone(self, editor: TextEditor) -> None:
        """Launch the debugger in standalone mode.

        Args:
            editor (TextEditor): the text editor"""

        self._launch_debugger(editor.path)

    def _launch_debugger(self, path: str) -> None:
        """Launch the debugger.

        Args:
            path (str): the file path"""

        self.manager.latest = self
        self.launch(path)

    def launch_config(self, config: dict[str, str]) -> None:
        """Launch the debugger with the given configuration.

        Args:
            config (dict[str, str]): the configuration"""

        self.manager.latest = self
        self.launch(
            self.format_path(config["program"]),
            self.format_path(config.get("cwd", "")),
        )

    def launch(self, path: str, cwd: str = None) -> None:
        """Debug the file.

        This method should be implemented by the subclass.

        Args:
            path (str): the file path
            cwd (str): the working directory"""

        raise NotImplementedError

    def update_breakpoints(self, file_path: str, line_numbers: set[int]) -> None:
        """Update the breakpoints for the file.

        Args:
            file_path (str): the file path
            line_numbers (set[int]): the set of line numbers"""

        self.breakpoints[file_path] = line_numbers

    def stop(self):
        """Stop the debugger."""

        raise NotImplementedError

    def restart(self):
        """Restart the debugger."""

        raise NotImplementedError

    def step_in(self) -> None:
        """Step through the code."""

        raise NotImplementedError

    def step_over(self) -> None:
        """Step over the code."""

        raise NotImplementedError

    def step_out(self) -> None:
        """Step out of the code."""

        raise NotImplementedError

    def continue_pause(self) -> None:
        """Toggle between continue and pause."""

        raise NotImplementedError

    def set_variable(self, name: str, value: str) -> None:
        """Set the value of a local variable.

        Args:
            name (str): local variable name
            value (str): the value"""

        raise NotImplementedError



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Menubutton

if typing.TYPE_CHECKING:
    from .menu import Menu


class Command(Menubutton):
    """A menu item

    Inherits from Menubutton"""

    def __init__(self, master, text, command=lambda *_: ..., *args, **kwargs) -> None:
        """Create a menu item

        Args:
            master: The parent widget
            text: The text to display on the menu item
            command: The command to run when the item is clicked
            *args: Additional arguments to pass to the Menubutton
        """

        super().__init__(master, *args, **kwargs)
        self.command = command
        self.menu = master.master

        self.config(
            text=text,
            anchor=tk.W,
            font=self.base.settings.uifont,
            cursor="hand2",
            padx=20,
            pady=2,
            **self.base.theme.menu.item,
        )
        self.bind("<Button-1>", self.on_click)

    def on_click(self, *_) -> None:
        self.menu.event_chosen()
        self.command()



import tkinter as tk


class FindResults(tk.Label):
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base

        self.init()
        self.config(
            padx=1,
            width=10,
            font=("Helvetica", 10),
            **self.base.theme.findreplace.label,
        )

    def show(self, n):
        if not n:
            self.config(text="No results")
            self.config(fg="#f48771")
        else:
            self.config(text=f"{n} results")
            self.config(fg=self.base.theme.findreplace.label.foreground)

    def init(self):
        self.config(text="No results")
        self.config(fg=self.base.theme.editors.foreground)



from __future__ import annotations

import os
import typing

from biscuit.common import is_image

from ..git.diff import DiffEditor
from .html import HTMLEditor
from .image import ImageViewer
from .markdown import MDEditor
from .text import TextEditor

if typing.TYPE_CHECKING:
    from .editorbase import BaseEditor


class EditorTypes:
    def __init__(self, base):
        self.base = base
        self._types = {}

    def register(self, editor_type: BaseEditor):
        self._types[editor_type.name] = editor_type

    def get(self, name):
        return self._types.get(name)

    def get_all(self):
        return self._types.values()

    def get_names(self):
        return self._types.keys()

    def get_default(self):
        return self.get("text")

    def get_editor(
        self,
        master,
        path="",
        exists=True,
        path2="",
        diff=False,
        language="",
        load_file=True,
        standalone=False,
    ) -> TextEditor | DiffEditor | MDEditor | ImageViewer:
        """Get the suitable editor based on the path, exists, diff values passed.

        Args:
            base: The parent widget
            path (str): The path of the file to be opened
            exists (bool): Whether the file exists
            path2 (str): The path of the file to be opened in diff, required if diff=True is passed
            diff (bool): Whether the file is to be opened in diff editor
            language (str): The language of the file

        Returns:
            TextEditor | DiffEditor | MDEditor | ImageViewer:
                The suitable editor based on the path, exists, diff values passed"""

        if diff:
            return DiffEditor(master, path, exists, path2, standalone=standalone)

        if path and os.path.isfile(path):
            if is_image(path):
                return ImageViewer(master, path)
            if any(
                path.endswith(i) for i in (".md", ".markdown", ".mdown", ".rst", ".mkd")
            ):
                return MDEditor(master, path, exists=exists)
            if path.endswith(".html") or path.endswith(".htm"):
                return HTMLEditor(master, path, exists=exists)

            return TextEditor(
                master, path, exists, language=language, load_file=load_file
            )

        return TextEditor(master, exists=exists, language=language, load_file=False)



import tkinter as tk

from .native import Frame


class KeyB(Frame):
    """Visual representation of a key binding.

    Args:
        master (tk.Widget): The parent widget.
        key (str): The key to be displayed."""

    def __init__(self, master, key: str, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.key = key
        self.config(bg=self.base.theme.border)

        frame = Frame(self, bg=self.base.theme.editors.background)
        frame.pack(pady=1, padx=1)

        self.label = tk.Label(
            frame,
            text=key,
            bg=self.base.theme.border,  # self.base.theme.editors.background,
            fg=self.base.theme.editors.foreground,  # self.base.theme.border,
            font=("Consolas", 8),
        )
        self.label.pack(pady=(1, 3), padx=1, side=tk.LEFT)


class Shortcut(Frame):
    """Visual representation of a shortcut key combination.

    Args:
        master (tk.Widget): The parent widget.
        shortcuts (tuple[str]): Tuple of shortcuts to be displayed."""

    def __init__(self, master, keys: tuple[str], *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self._keys = keys

        self.labels = []
        self.bg, self.fg, _, _ = self.base.theme.editors.values()
        self.render()

    def render(self) -> None:
        """Render the widget."""

        for key in self._keys[:-1]:
            self.add_key(key)
            self.add_separator()
        self.add_key(self._keys[-1])

    def add_separator(self) -> None:
        """Add a '+' separator between shortcuts."""

        l = tk.Label(
            self, text="+", bg=self.bg, fg=self.base.theme.border, font=("Consolas", 7)
        )
        l.pack(padx=2, side=tk.LEFT)
        self.labels.append(l)

    def add_key(self, key: str) -> None:
        """Add a shortcut to the widget."""

        l = KeyB(self, key)
        l.pack(padx=2, side=tk.LEFT)
        self.labels.append(l)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .. import Settings


class Bindings:
    """Class for managing key bindings"""

    def __init__(self, master: Settings) -> None:
        self.base = master.base

        self.new_file = "<Control-n>"
        self.new_window = "<Control-N>"
        self.open_file = "<Control-o>"
        self.open_dir = "<Control-O>"
        self.save = "<Control-s>"
        self.save_as = "<Control-S>"
        self.close_file = "<Control-w>"
        self.goto_line = "<Control-g>"
        self.quit = "<Control-q>"
        self.undo = "<Control-z>"
        self.redo = "<Control-y>"
        self.restore_closed_tab = "<Control-T>"
        self.close_all_tabs = "<Control-W>"
        self.change_tab = "<Control-Tab>"
        self.change_tab_back = "<Control-Shift-Tab>"
        self.split_tab = "<Control-\\>"

        self.command_palette = "<Control-P>"
        self.file_search = "<Control-p>"
        self.symbol_outline = "<Control-j>"

        self.panel = "<Control-grave>"
        self.sidebar = "<Control-b>"
        self.secondary_sidebar = "<Control-B>"
        self.directory_tree = "<Control-E>"
        self.extensions = "<Control-X>"
        self.global_search = "<Control-F>"
        self.debugger = "<Control-D>"
        self.git = "<Control-G>"
        self.assistant = "<Control-A>"
        self.logs = "<Control-U>"

        self.open_settings = "<Control-comma>"
        self.restore_recent_session = "<Control-Alt-r>"
        self.open_recent_folders = "<Control-r>"
        self.open_recent_files = "<Control-R>"
        self.open_recent_session = "<Control-Alt-r>"
